"""
VTool - Computer vision tools

Autogenerate Command:
    mkinit vtool_ibeis -i
"""
# flake8: noqa
__version__ = '2.3.0'
__author__ = 'Jon Crall, Avi Weinstock, Chuck Stewart, Hendrik Weideman, Jason Parham, Zackary Rutfield'
__author_email__ = 'erotemic@gmail.com'
__url__ = 'https://github.com/Erotemic/vtool_ibeis'



__submodules__ = [
    'histogram',
    'features',
    'linalg',
    'blend',
    'image_shared',
    'image',
    'exif',
    'distance',
    'keypoint',
    'ellipse',
    'patch',
    'chip',
    'spatial_verification',
    'trig',
    'util_math',
    'matching',
    'geometry',
    'nearest_neighbors',
    'clustering2',
    'other',
    'numpy_utils',
    'confusion',
    'score_normalization',
    'symbolic',
    'demodata',
]

from vtool_ibeis import histogram
from vtool_ibeis import features
from vtool_ibeis import linalg
from vtool_ibeis import blend
from vtool_ibeis import image_shared
from vtool_ibeis import image
from vtool_ibeis import exif
from vtool_ibeis import distance
from vtool_ibeis import keypoint
from vtool_ibeis import ellipse
from vtool_ibeis import patch
from vtool_ibeis import chip
from vtool_ibeis import spatial_verification
from vtool_ibeis import trig
from vtool_ibeis import util_math
from vtool_ibeis import matching
from vtool_ibeis import geometry
from vtool_ibeis import nearest_neighbors
from vtool_ibeis import clustering2
from vtool_ibeis import other
from vtool_ibeis import numpy_utils
from vtool_ibeis import confusion
from vtool_ibeis import score_normalization
from vtool_ibeis import symbolic
from vtool_ibeis import demodata

# TODO: de-utoolificaiton: replace utool with ubelt
from vtool_ibeis import histogram as htool
from vtool_ibeis import linalg as ltool
from vtool_ibeis import image as gtool
from vtool_ibeis import exif as exiftool
from vtool_ibeis import keypoint as ktool
from vtool_ibeis import ellipse as etool
from vtool_ibeis import patch as ptool
from vtool_ibeis import chip as ctool
from vtool_ibeis import spatial_verification as svtool
from vtool_ibeis import clustering2 as clustertool
from vtool_ibeis import trig
from vtool_ibeis import util_math as mtool

import sys

"""
break
"""

# <AUTOGEN_INIT>
from vtool_ibeis import histogram
from vtool_ibeis import features
from vtool_ibeis import linalg
from vtool_ibeis import blend
from vtool_ibeis import image_shared
from vtool_ibeis import image
from vtool_ibeis import exif
from vtool_ibeis import distance
from vtool_ibeis import keypoint
from vtool_ibeis import ellipse
from vtool_ibeis import patch
from vtool_ibeis import chip
from vtool_ibeis import spatial_verification
from vtool_ibeis import trig
from vtool_ibeis import util_math
from vtool_ibeis import matching
from vtool_ibeis import geometry
from vtool_ibeis import nearest_neighbors
from vtool_ibeis import clustering2
from vtool_ibeis import other
from vtool_ibeis import numpy_utils
from vtool_ibeis import confusion
from vtool_ibeis import score_normalization
from vtool_ibeis import symbolic
from vtool_ibeis import demodata

from vtool_ibeis.histogram import (argsubextrema2, argsubmax, argsubmax2,
                             argsubmaxima, argsubmaxima2, argsubmin2,
                             argsubminima2, get_histinfo_str, hist_argmaxima,
                             hist_argmaxima2, hist_edges_to_centers,
                             interpolate_submaxima, interpolated_histogram,
                             linear_interpolation, maxima_neighbors,
                             maximum_parabola_point, show_hist_submaxima,
                             show_ori_image, show_ori_image_ondisk,
                             subbin_bounds, wrap_histogram,)
from vtool_ibeis.features import (detect_opencv_keypoints,
                            extract_feature_from_patch, extract_features,
                            get_extract_features_default_params, test_mser,)
from vtool_ibeis.linalg import (TRANSFORM_DTYPE, add_homogenous_coordinate,
                          affine_around_mat3x3, affine_mat3x3, det_ltri,
                          dot_ltri, gauss2d_pdf, inv_ltri, normalize,
                          normalize_rows, random_affine_args,
                          random_affine_transform,
                          remove_homogenous_coordinate,
                          rotation_around_bbox_mat3x3, rotation_around_mat3x3,
                          rotation_mat2x2, rotation_mat3x3,
                          scale_around_mat3x3, scale_mat3x3, shear_mat3x3, svd,
                          transform_around, transform_points_with_homography,
                          translation_mat3x3, whiten_xy_points,)
from vtool_ibeis.blend import (blend_images, blend_images_average,
                         blend_images_average_stack, blend_images_mult_average,
                         blend_images_multiply, ensure_alpha_channel,
                         ensure_grayscale, gamma_adjust,
                         gridsearch_addWeighted, gridsearch_image_function,
                         overlay_alpha_images, testdata_blend,)
from vtool_ibeis.image_shared import (open_pil_image, print_image_checks,)
from vtool_ibeis.image import (EXIF_TAG_DATETIME, EXIF_TAG_GPS, LINE_AA,
                         affine_warp_around_center, clipwhite,
                         clipwhite_ondisk, combine_offset_lists,
                         convert_colorspace, convert_image_list_colorspace,
                         crop_out_imgfill, cvt_BGR2L, cvt_BGR2RGB, draw_text,
                         embed_channels, embed_in_square_image,
                         ensure_3channel, ensure_4channel,
                         filterflags_valid_images, find_pixel_value_index,
                         get_num_channels, get_pixel_dist,
                         get_round_scaled_dsize, get_scale_factor, get_size,
                         imread, imread_remote_s3, imread_remote_url, imwrite,
                         imwrite_fallback, infer_vert,
                         make_channels_comparable, make_white_transparent,
                         montage, open_image_size, pad_image, pad_image_ondisk,
                         padded_resize, perlin_noise, rectify_to_float01,
                         rectify_to_square, rectify_to_uint8, resize,
                         resize_image_by_scale, resize_mask, resize_thumb,
                         resize_to_maxdims, resize_to_maxdims_ondisk,
                         resized_clamped_thumb_dims, resized_dims_and_ratio,
                         rotate_image, rotate_image_ondisk, shear,
                         stack_image_list, stack_image_list_special,
                         stack_image_recurse, stack_images, stack_multi_images,
                         stack_multi_images2, stack_square_images,
                         subpixel_values, testdata_imglist, warpAffine,
                         warpHomog,)
from vtool_ibeis.exif import (DATETIMEORIGINAL_TAGID, EXIF_TAG_TO_TAGID,
                        GPSDATE_CODE, GPSINFO_CODE, GPSLATITUDEREF_CODE,
                        GPSLATITUDE_CODE, GPSLONGITUDEREF_CODE,
                        GPSLONGITUDE_CODE, GPSTIME_CODE, GPS_TAG_TO_GPSID,
                        ORIENTATION_000, ORIENTATION_090, ORIENTATION_180,
                        ORIENTATION_270, ORIENTATION_CODE, ORIENTATION_DICT,
                        ORIENTATION_DICT_INVERSE, ORIENTATION_ORDER_LIST,
                        ORIENTATION_UNDEFINED, SENSITIVITYTYPE_CODE,
                        check_exif_keys, convert_degrees, get_exif_dict,
                        get_exif_dict2, get_exif_tagids, get_exist,
                        get_lat_lon, get_orientation, get_orientation_str,
                        get_unixtime, get_unixtime_gps,
                        make_exif_dict_human_readable, parse_exif_unixtime,
                        parse_exif_unixtime_gps, read_all_exif_tags, read_exif,
                        read_exif_tags, read_one_exif_tag,)
from vtool_ibeis.distance import (L1, L2, L2_root_sift, L2_sift, L2_sift_sqrd,
                            L2_sqrd, TEMP_VEC_DTYPE, VALID_DISTS, bar_L2_sift,
                            bar_cos_sift, closest_point, compute_distances,
                            cos_sift, cosine_dist, cyclic_distance,
                            det_distance, emd, haversine, hist_isect,
                            nearest_point, ori_distance, pdist_argsort,
                            pdist_indicies, safe_pdist, signed_cyclic_distance,
                            signed_ori_distance, testdata_hist, testdata_sift2,
                            understanding_pseudomax_props, wrapped_distance,)
from vtool_ibeis.keypoint import (GRAVITY_THETA, KPTS_DTYPE, LOC_DIMS, ORI_DIM,
                            SCAX_DIM, SCAY_DIM, SHAPE_DIMS, SKEW_DIM, XDIM,
                            YDIM, augment_2x2_with_translation, cast_split,
                            convert_kptsZ_to_kpts, decompose_Z_to_RV_mats2x2,
                            decompose_Z_to_V_2x2, decompose_Z_to_invV_2x2,
                            decompose_Z_to_invV_mats2x2,
                            flatten_invV_mats_to_kpts, get_RV_mats2x2,
                            get_RV_mats_3x3, get_V_mats, get_Z_mats,
                            get_even_point_sample, get_grid_kpts,
                            get_invVR_mats2x2, get_invVR_mats3x3,
                            get_invVR_mats_oris, get_invVR_mats_shape,
                            get_invVR_mats_sqrd_scale, get_invVR_mats_xys,
                            get_invV_mats, get_invV_mats2x2, get_invV_mats3x3,
                            get_invVs, get_kpts_dlen_sqrd,
                            get_kpts_eccentricity, get_kpts_image_extent,
                            get_kpts_strs, get_kpts_wh,
                            get_match_spatial_squared_error, get_ori_mats,
                            get_ori_strs, get_oris, get_scales, get_shape_strs,
                            get_sqrd_scales,
                            get_transforms_from_patch_image_kpts,
                            get_uneven_point_sample, get_xy_strs, get_xys,
                            invert_invV_mats, kp_cpp_infostr, kpts_docrepr,
                            kpts_repr, offset_kpts, rectify_invV_mats_are_up,
                            transform_kpts, transform_kpts_to_imgspace,
                            transform_kpts_xys,)
from vtool_ibeis.ellipse import (adaptive_scale, check_kpts_in_bounds,
                           circular_distance, expand_kpts, expand_scales,
                           expand_subscales, extrema_neighbors, find_maxima,
                           find_maxima_with_neighbors, gradient_magnitude,
                           homogenous_circle_pts, interpolate_between,
                           interpolate_maxima, interpolate_peaks,
                           interpolate_peaks2, kpts_matrices,
                           sample_ell_border_pts, sample_ell_border_vals,
                           sample_uniform, subscale_peaks,)
from vtool_ibeis.patch import (GaussianBlurInplace, draw_kp_ori_steps,
                         find_dominant_kp_orientations, find_kpts_direction,
                         find_patch_dominant_orientations,
                         gaussian_average_patch, gaussian_patch,
                         gaussian_weight_patch, generate_to_patch_transforms,
                         get_cross_patch, get_no_symbol,
                         get_orientation_histogram, get_star2_patch,
                         get_star_patch, get_stripe_patch, get_test_patch,
                         get_unwarped_patch, get_unwarped_patches,
                         get_warped_patch, get_warped_patches, gradient_fill,
                         intern_warp_single_patch, inverted_sift_patch,
                         make_test_image_keypoints,
                         patch_gaussian_weighted_average_intensities,
                         patch_gradient, patch_mag, patch_ori,
                         show_gaussian_patch,
                         show_patch_orientation_estimation,
                         test_ondisk_find_patch_fpath_dominant_orientations,
                         test_show_gaussian_patches,
                         test_show_gaussian_patches2, testdata_patch,)
from vtool_ibeis.chip import (ScaleStrat, apply_filter_funcs, compute_chip,
                        extract_chip_from_gpath,
                        extract_chip_from_gpath_into_square,
                        extract_chip_from_img, extract_chip_into_square,
                        get_extramargin_measures, get_image_to_chip_transform,
                        get_scaled_size_with_dlen, gridsearch_chipextract,
                        testshow_extramargin_info,)
from vtool_ibeis.spatial_verification import (HAVE_SVER_C_WRAPPER, INDEX_DTYPE,
                                        SV_DTYPE, VERBOSE_SVER,
                                        build_affine_lstsqrs_Mx6,
                                        build_lstsqrs_Mx9, compute_affine,
                                        compute_homog,
                                        estimate_refined_transform,
                                        get_affine_inliers,
                                        get_best_affine_inliers,
                                        get_best_affine_inliers_,
                                        get_normalized_affine_inliers,
                                        refine_inliers, spatially_verify_kpts,
                                        test_affine_errors, test_homog_errors,
                                        testdata_matching_affine_inliers,
                                        testdata_matching_affine_inliers_normalized,
                                        try_svd, unnormalize_transform,)
from vtool_ibeis.trig import (atan2,)
from vtool_ibeis.util_math import (TAU, beaton_tukey_loss, beaton_tukey_weight,
                             breakup_equal_streak, ensure_monotone_decreasing,
                             ensure_monotone_increasing,
                             ensure_monotone_strictly_decreasing,
                             ensure_monotone_strictly_increasing, eps,
                             gauss_func1d, gauss_func1d_unnormalized,
                             gauss_parzen_est, group_consecutive, iceil,
                             interpolate_nans, iround, logistic_01, logit,
                             non_decreasing, non_increasing,
                             strictly_decreasing, strictly_increasing,
                             test_language_modulus,)
from vtool_ibeis.matching import (AnnotPairFeatInfo, AssignTup, MatchingError,
                            NORM_CHIP_CONFIG, PSEUDO_MAX_DIST,
                            PSEUDO_MAX_DIST_SQRD, PSEUDO_MAX_VEC_COMPONENT,
                            PairwiseMatch, SUM_OPS, VSONE_ASSIGN_CONFIG,
                            VSONE_DEFAULT_CONFIG, VSONE_FEAT_CONFIG,
                            VSONE_PI_DICT, VSONE_RATIO_CONFIG,
                            VSONE_SVER_CONFIG, assign_symmetric_matches,
                            assign_unconstrained_matches,
                            asymmetric_correspondence, csum, demodata_match,
                            empty_assign, empty_neighbors,
                            ensure_metadata_dlen_sqrd, ensure_metadata_feats,
                            ensure_metadata_flann, ensure_metadata_normxy,
                            ensure_metadata_vsone, flag_sym_slow,
                            flag_symmetric_matches, invsum,
                            normalized_nearest_neighbors,
                            symmetric_correspondence, testdata_annot_metadata,)
from vtool_ibeis.geometry import (bbox_center, bbox_from_center_wh, bbox_from_extent,
                            bbox_from_verts, bbox_from_xywh,
                            bboxes_from_vert_list, closest_point_on_bbox,
                            closest_point_on_line,
                            closest_point_on_line_segment,
                            closest_point_on_vert_segments,
                            cvt_bbox_xywh_to_pt1pt2, distance_to_lineseg,
                            draw_border, draw_verts, extent_from_bbox,
                            extent_from_verts, get_pointset_extent_wh,
                            get_pointset_extents, point_inside_bbox,
                            scale_bbox, scale_extents, scaled_verts_from_bbox,
                            scaled_verts_from_bbox_gen, union_extents,
                            verts_from_bbox, verts_list_from_bboxes_list,)
from vtool_ibeis.nearest_neighbors import (AnnoyWrapper, ann_flann_once,
                                     assign_to_centroids, flann_augment,
                                     flann_cache, flann_index_time_experiment,
                                     get_flann_cfgstr, get_flann_fpath,
                                     get_flann_params, get_flann_params_cfgstr,
                                     get_kdtree_flann_params, invertible_stack,
                                     test_annoy, test_cv2_flann, tune_flann,)
from vtool_ibeis.clustering2 import (AnnoyWraper, apply_grouping, apply_grouping_,
                               apply_grouping_iter, apply_grouping_iter2,
                               apply_jagged_grouping, example_binary,
                               find_duplicate_items, group_indices, groupby,
                               groupby_dict, groupby_gen, groupedzip,
                               invert_apply_grouping, invert_apply_grouping2,
                               invert_apply_grouping3, jagged_group,
                               plot_centroids, sorted_indices_ranges,
                               tune_flann2, uniform_sample_hypersphere,
                               unsupervised_multicut_labeling,)
from vtool_ibeis.other import (and_lists, argsort_groups, argsort_records,
                         assert_zipcompress, asserteq, atleast_3channels,
                         atleast_nd, atleast_shape,
                         calc_error_bars_from_sample,
                         calc_sample_from_error_bars, check_sift_validity,
                         clipnorm, colwise_operation, compare_implementations,
                         compare_matrix_columns, compare_matrix_to_rows,
                         componentwise_dot, compress2,
                         compute_ndarray_unique_rowids_unsafe,
                         compute_unique_arr_dataids, compute_unique_data_ids,
                         compute_unique_data_ids_,
                         compute_unique_integer_data_ids, ensure_rng,
                         ensure_shape, find_best_undirected_edge_indexes,
                         find_elbow_point, find_first_true_indices,
                         find_k_true_indicies, find_next_true_indices,
                         flag_intersection, fromiter_nd, get_covered_mask,
                         get_crop_slices, get_uncovered_mask,
                         get_undirected_edge_ids, grab_webcam_image,
                         greedy_setcover, inbounds, index_partition,
                         intersect1d_reduce, intersect2d_flags,
                         intersect2d_indices, intersect2d_numpy,
                         intersect2d_structured_numpy, iter_reduce_ufunc,
                         list_compress_, list_take_, make_video, make_video2,
                         median_abs_dev, mult_lists, multigroup_lookup,
                         multigroup_lookup_naive, nearest_point,
                         # nonunique_row_flags,
                         # nonunique_row_indexes,
                               norm01,
                         or_lists, pad_vstack, rebuild_partition,
                         rowwise_operation, safe_argmax, safe_cat, safe_div,
                         safe_extreme, safe_max, safe_min, safe_vstack,
                         significant_shape, structure_rows, take2,
                         take_col_per_row, to_undirected_edges, trytake,
                         unique_rows, unstructure_rows,
                         weighted_average_scoring, weighted_geometic_mean,
                         weighted_geometic_mean_unnormalized, zipcat,
                         zipcompress, zipcompress_safe, ziptake, zstar_value,)
from vtool_ibeis.numpy_utils import (atleast_nd, ensure_shape, fromiter_nd,
                               index_to_boolmask, iter_reduce_ufunc,
                               multiaxis_reduce, unique_row_indexes,)
from vtool_ibeis.confusion import (ConfusionMetrics, draw_precision_recall_curve,
                             draw_roc_curve, interact_roc_factory,
                             interpolate_precision_recall,
                             interpolate_replbounds, nan_to_num,
                             testdata_scores_labels,)
from vtool_ibeis.score_normalization import (ScoreNormVisualizeClass,
                                       ScoreNormalizer, check_unused_kwargs,
                                       estimate_pdf, find_clip_range,
                                       flatten_scores, get_left_area,
                                       get_right_area, inspect_pdfs,
                                       learn_score_normalization,
                                       normalize_scores, partition_scores,
                                       plot_postbayes_pdf, plot_prebayes_pdf,
                                       test_score_normalization,
                                       testdata_score_normalier,)
from vtool_ibeis.symbolic import (check_expr_eq, custom_sympy_attrs, evalprint,
                            symbolic_randcheck, sympy_latex_repr, sympy_mat,
                            sympy_numpy_repr,)
from vtool_ibeis.demodata import (DEFAULT_DTYPE, dummy_img, dummy_seed,
                            force_kpts_feasibility, get_dummy_dpts,
                            get_dummy_invV_mats, get_dummy_kpts,
                            get_dummy_kpts_pair, get_dummy_matching_kpts,
                            get_dummy_xy, get_kpts_dummy_img,
                            get_testdata_kpts, make_dummy_fm, perterb_kpts,
                            perterbed_grid_kpts, testdata_binary_scores,
                            testdata_dummy_matches, testdata_dummy_sift,
                            testdata_nonmonotonic, testdata_ratio_matches,)

__all__ = ['AnnotPairFeatInfo', 'AnnoyWraper', 'AnnoyWrapper', 'AssignTup',
           'ConfusionMetrics', 'DATETIMEORIGINAL_TAGID', 'DEFAULT_DTYPE',
           'EXIF_TAG_DATETIME', 'EXIF_TAG_GPS', 'EXIF_TAG_TO_TAGID',
           'GPSDATE_CODE', 'GPSINFO_CODE', 'GPSLATITUDEREF_CODE',
           'GPSLATITUDE_CODE', 'GPSLONGITUDEREF_CODE', 'GPSLONGITUDE_CODE',
           'GPSTIME_CODE', 'GPS_TAG_TO_GPSID', 'GRAVITY_THETA',
           'GaussianBlurInplace', 'HAVE_SVER_C_WRAPPER', 'INDEX_DTYPE',
           'KPTS_DTYPE', 'L1', 'L2', 'L2_root_sift', 'L2_sift', 'L2_sift_sqrd',
           'L2_sqrd', 'LINE_AA', 'LOC_DIMS', 'MatchingError',
           'NORM_CHIP_CONFIG', 'ORIENTATION_000', 'ORIENTATION_090',
           'ORIENTATION_180', 'ORIENTATION_270', 'ORIENTATION_CODE',
           'ORIENTATION_DICT', 'ORIENTATION_DICT_INVERSE',
           'ORIENTATION_ORDER_LIST', 'ORIENTATION_UNDEFINED', 'ORI_DIM',
           'PSEUDO_MAX_DIST', 'PSEUDO_MAX_DIST_SQRD',
           'PSEUDO_MAX_VEC_COMPONENT', 'PairwiseMatch', 'SCAX_DIM', 'SCAY_DIM',
           'SENSITIVITYTYPE_CODE', 'SHAPE_DIMS', 'SKEW_DIM', 'SUM_OPS',
           'SV_DTYPE', 'ScaleStrat', 'ScoreNormVisualizeClass',
           'ScoreNormalizer', 'TAU', 'TEMP_VEC_DTYPE', 'TRANSFORM_DTYPE',
           'VALID_DISTS', 'VERBOSE_SVER', 'VSONE_ASSIGN_CONFIG',
           'VSONE_DEFAULT_CONFIG', 'VSONE_FEAT_CONFIG', 'VSONE_PI_DICT',
           'VSONE_RATIO_CONFIG', 'VSONE_SVER_CONFIG', 'XDIM', 'YDIM',
           'adaptive_scale', 'add_homogenous_coordinate',
           'affine_around_mat3x3', 'affine_mat3x3',
           'affine_warp_around_center', 'and_lists', 'ann_flann_once',
           'apply_filter_funcs', 'apply_grouping', 'apply_grouping_',
           'apply_grouping_iter', 'apply_grouping_iter2',
           'apply_jagged_grouping', 'argsort_groups', 'argsort_records',
           'argsubextrema2', 'argsubmax', 'argsubmax2', 'argsubmaxima',
           'argsubmaxima2', 'argsubmin2', 'argsubminima2',
           'assert_zipcompress', 'asserteq', 'assign_symmetric_matches',
           'assign_to_centroids', 'assign_unconstrained_matches',
           'asymmetric_correspondence', 'atan2', 'atleast_3channels',
           'atleast_nd', 'atleast_nd', 'atleast_shape',
           'augment_2x2_with_translation', 'bar_L2_sift', 'bar_cos_sift',
           'bbox_center', 'bbox_from_center_wh', 'bbox_from_extent',
           'bbox_from_verts', 'bbox_from_xywh', 'bboxes_from_vert_list',
           'beaton_tukey_loss', 'beaton_tukey_weight', 'blend', 'blend_images',
           'blend_images_average', 'blend_images_average_stack',
           'blend_images_mult_average', 'blend_images_multiply',
           'breakup_equal_streak', 'build_affine_lstsqrs_Mx6',
           'build_lstsqrs_Mx9', 'calc_error_bars_from_sample',
           'calc_sample_from_error_bars', 'cast_split', 'check_exif_keys',
           'check_expr_eq', 'check_kpts_in_bounds', 'check_sift_validity',
           'check_unused_kwargs', 'chip', 'circular_distance', 'clipnorm',
           'clipwhite', 'clipwhite_ondisk', 'closest_point',
           'closest_point_on_bbox', 'closest_point_on_line',
           'closest_point_on_line_segment', 'closest_point_on_vert_segments',
           'clustering2', 'colwise_operation', 'combine_offset_lists',
           'compare_implementations', 'compare_matrix_columns',
           'compare_matrix_to_rows', 'componentwise_dot', 'compress2',
           'compute_affine', 'compute_chip', 'compute_distances',
           'compute_homog', 'compute_ndarray_unique_rowids_unsafe',
           'compute_unique_arr_dataids', 'compute_unique_data_ids',
           'compute_unique_data_ids_', 'compute_unique_integer_data_ids',
           'confusion', 'convert_colorspace', 'convert_degrees',
           'convert_image_list_colorspace', 'convert_kptsZ_to_kpts',
           'cos_sift', 'cosine_dist', 'crop_out_imgfill', 'csum',
           'custom_sympy_attrs', 'cvt_BGR2L', 'cvt_BGR2RGB',
           'cvt_bbox_xywh_to_pt1pt2', 'cyclic_distance',
           'decompose_Z_to_RV_mats2x2', 'decompose_Z_to_V_2x2',
           'decompose_Z_to_invV_2x2', 'decompose_Z_to_invV_mats2x2',
           'demodata', 'demodata_match', 'det_distance', 'det_ltri',
           'detect_opencv_keypoints', 'distance', 'distance_to_lineseg',
           'dot_ltri', 'draw_border', 'draw_kp_ori_steps',
           'draw_precision_recall_curve', 'draw_roc_curve', 'draw_text',
           'draw_verts', 'dummy_img', 'dummy_seed', 'ellipse',
           'embed_channels', 'embed_in_square_image', 'emd', 'empty_assign',
           'empty_neighbors', 'ensure_3channel', 'ensure_4channel',
           'ensure_alpha_channel', 'ensure_grayscale',
           'ensure_metadata_dlen_sqrd', 'ensure_metadata_feats',
           'ensure_metadata_flann', 'ensure_metadata_normxy',
           'ensure_metadata_vsone', 'ensure_monotone_decreasing',
           'ensure_monotone_increasing', 'ensure_monotone_strictly_decreasing',
           'ensure_monotone_strictly_increasing', 'ensure_rng', 'ensure_shape',
           'ensure_shape', 'eps', 'estimate_pdf', 'estimate_refined_transform',
           'evalprint', 'example_binary', 'exif', 'expand_kpts',
           'expand_scales', 'expand_subscales', 'extent_from_bbox',
           'extent_from_verts', 'extract_chip_from_gpath',
           'extract_chip_from_gpath_into_square', 'extract_chip_from_img',
           'extract_chip_into_square', 'extract_feature_from_patch',
           'extract_features', 'extrema_neighbors', 'features',
           'filterflags_valid_images', 'find_best_undirected_edge_indexes',
           'find_clip_range', 'find_dominant_kp_orientations',
           'find_duplicate_items', 'find_elbow_point',
           'find_first_true_indices', 'find_k_true_indicies',
           'find_kpts_direction', 'find_maxima', 'find_maxima_with_neighbors',
           'find_next_true_indices', 'find_patch_dominant_orientations',
           'find_pixel_value_index', 'flag_intersection', 'flag_sym_slow',
           'flag_symmetric_matches', 'flann_augment', 'flann_cache',
           'flann_index_time_experiment', 'flatten_invV_mats_to_kpts',
           'flatten_scores', 'force_kpts_feasibility', 'fromiter_nd',
           'fromiter_nd', 'gamma_adjust', 'gauss2d_pdf', 'gauss_func1d',
           'gauss_func1d_unnormalized', 'gauss_parzen_est',
           'gaussian_average_patch', 'gaussian_patch', 'gaussian_weight_patch',
           'generate_to_patch_transforms', 'geometry', 'get_RV_mats2x2',
           'get_RV_mats_3x3', 'get_V_mats', 'get_Z_mats', 'get_affine_inliers',
           'get_best_affine_inliers', 'get_best_affine_inliers_',
           'get_covered_mask', 'get_crop_slices', 'get_cross_patch',
           'get_dummy_dpts', 'get_dummy_invV_mats', 'get_dummy_kpts',
           'get_dummy_kpts_pair', 'get_dummy_matching_kpts', 'get_dummy_xy',
           'get_even_point_sample', 'get_exif_dict', 'get_exif_dict2',
           'get_exif_tagids', 'get_exist',
           'get_extract_features_default_params', 'get_extramargin_measures',
           'get_flann_cfgstr', 'get_flann_fpath', 'get_flann_params',
           'get_flann_params_cfgstr', 'get_grid_kpts', 'get_histinfo_str',
           'get_image_to_chip_transform', 'get_invVR_mats2x2',
           'get_invVR_mats3x3', 'get_invVR_mats_oris', 'get_invVR_mats_shape',
           'get_invVR_mats_sqrd_scale', 'get_invVR_mats_xys', 'get_invV_mats',
           'get_invV_mats2x2', 'get_invV_mats3x3', 'get_invVs',
           'get_kdtree_flann_params', 'get_kpts_dlen_sqrd',
           'get_kpts_dummy_img', 'get_kpts_eccentricity',
           'get_kpts_image_extent', 'get_kpts_strs', 'get_kpts_wh',
           'get_lat_lon', 'get_left_area', 'get_match_spatial_squared_error',
           'get_no_symbol', 'get_normalized_affine_inliers',
           'get_num_channels', 'get_ori_mats', 'get_ori_strs',
           'get_orientation', 'get_orientation_histogram',
           'get_orientation_str', 'get_oris', 'get_pixel_dist',
           'get_pointset_extent_wh', 'get_pointset_extents', 'get_right_area',
           'get_round_scaled_dsize', 'get_scale_factor',
           'get_scaled_size_with_dlen', 'get_scales', 'get_shape_strs',
           'get_size', 'get_sqrd_scales', 'get_star2_patch', 'get_star_patch',
           'get_stripe_patch', 'get_test_patch', 'get_testdata_kpts',
           'get_transforms_from_patch_image_kpts', 'get_uncovered_mask',
           'get_undirected_edge_ids', 'get_uneven_point_sample',
           'get_unixtime', 'get_unixtime_gps', 'get_unwarped_patch',
           'get_unwarped_patches', 'get_warped_patch', 'get_warped_patches',
           'get_xy_strs', 'get_xys', 'grab_webcam_image', 'gradient_fill',
           'gradient_magnitude', 'greedy_setcover', 'gridsearch_addWeighted',
           'gridsearch_chipextract', 'gridsearch_image_function',
           'group_consecutive', 'group_indices', 'groupby', 'groupby_dict',
           'groupby_gen', 'groupedzip', 'haversine', 'hist_argmaxima',
           'hist_argmaxima2', 'hist_edges_to_centers', 'hist_isect',
           'histogram', 'homogenous_circle_pts', 'iceil', 'image',
           'image_shared', 'imread', 'imread_remote_s3', 'imread_remote_url',
           'imwrite', 'imwrite_fallback', 'inbounds', 'index_partition',
           'index_to_boolmask', 'infer_vert', 'inspect_pdfs',
           'interact_roc_factory', 'intern_warp_single_patch',
           'interpolate_between', 'interpolate_maxima', 'interpolate_nans',
           'interpolate_peaks', 'interpolate_peaks2',
           'interpolate_precision_recall', 'interpolate_replbounds',
           'interpolate_submaxima', 'interpolated_histogram',
           'intersect1d_reduce', 'intersect2d_flags', 'intersect2d_indices',
           'intersect2d_numpy', 'intersect2d_structured_numpy', 'inv_ltri',
           'invert_apply_grouping', 'invert_apply_grouping2',
           'invert_apply_grouping3', 'invert_invV_mats', 'inverted_sift_patch',
           'invertible_stack', 'invsum', 'iround', 'iter_reduce_ufunc',
           'iter_reduce_ufunc', 'jagged_group', 'keypoint', 'kp_cpp_infostr',
           'kpts_docrepr', 'kpts_matrices', 'kpts_repr',
           'learn_score_normalization', 'linalg', 'linear_interpolation',
           'list_compress_', 'list_take_', 'logistic_01', 'logit',
           'make_channels_comparable', 'make_dummy_fm',
           'make_exif_dict_human_readable', 'make_test_image_keypoints',
           'make_video', 'make_video2', 'make_white_transparent', 'matching',
           'maxima_neighbors', 'maximum_parabola_point', 'median_abs_dev',
           'montage', 'mult_lists', 'multiaxis_reduce', 'multigroup_lookup',
           'multigroup_lookup_naive', 'nan_to_num', 'nearest_neighbors',
           'nearest_point', 'nearest_point', 'non_decreasing',
           'non_increasing',
           # 'nonunique_row_flags',
           # 'nonunique_row_indexes',
           'norm01', 'normalize', 'normalize_rows', 'normalize_scores',
           'normalized_nearest_neighbors', 'numpy_utils', 'offset_kpts',
           'open_image_size', 'open_pil_image', 'or_lists', 'ori_distance',
           'other', 'overlay_alpha_images', 'pad_image', 'pad_image_ondisk',
           'pad_vstack', 'padded_resize', 'parse_exif_unixtime',
           'parse_exif_unixtime_gps', 'partition_scores', 'patch',
           'patch_gaussian_weighted_average_intensities', 'patch_gradient',
           'patch_mag', 'patch_ori', 'pdist_argsort', 'pdist_indicies',
           'perlin_noise', 'perterb_kpts', 'perterbed_grid_kpts',
           'plot_centroids', 'plot_postbayes_pdf', 'plot_prebayes_pdf',
           'point_inside_bbox', 'print_image_checks', 'random_affine_args',
           'random_affine_transform', 'read_all_exif_tags', 'read_exif',
           'read_exif_tags', 'read_one_exif_tag', 'rebuild_partition',
           'rectify_invV_mats_are_up', 'rectify_to_float01',
           'rectify_to_square', 'rectify_to_uint8', 'refine_inliers',
           'remove_homogenous_coordinate', 'resize', 'resize_image_by_scale',
           'resize_mask', 'resize_thumb', 'resize_to_maxdims',
           'resize_to_maxdims_ondisk', 'resized_clamped_thumb_dims',
           'resized_dims_and_ratio', 'rotate_image', 'rotate_image_ondisk',
           'rotation_around_bbox_mat3x3', 'rotation_around_mat3x3',
           'rotation_mat2x2', 'rotation_mat3x3', 'rowwise_operation',
           'safe_argmax', 'safe_cat', 'safe_div', 'safe_extreme', 'safe_max',
           'safe_min', 'safe_pdist', 'safe_vstack', 'sample_ell_border_pts',
           'sample_ell_border_vals', 'sample_uniform', 'scale_around_mat3x3',
           'scale_bbox', 'scale_extents', 'scale_mat3x3',
           'scaled_verts_from_bbox', 'scaled_verts_from_bbox_gen',
           'score_normalization', 'shear', 'shear_mat3x3',
           'show_gaussian_patch', 'show_hist_submaxima', 'show_ori_image',
           'show_ori_image_ondisk', 'show_patch_orientation_estimation',
           'signed_cyclic_distance', 'signed_ori_distance',
           'significant_shape', 'sorted_indices_ranges',
           'spatial_verification', 'spatially_verify_kpts', 'stack_image_list',
           'stack_image_list_special', 'stack_image_recurse', 'stack_images',
           'stack_multi_images', 'stack_multi_images2', 'stack_square_images',
           'strictly_decreasing', 'strictly_increasing', 'structure_rows',
           'subbin_bounds', 'subpixel_values', 'subscale_peaks', 'svd',
           'symbolic', 'symbolic_randcheck', 'symmetric_correspondence',
           'sympy_latex_repr', 'sympy_mat', 'sympy_numpy_repr', 'take2',
           'take_col_per_row', 'test_affine_errors', 'test_annoy',
           'test_cv2_flann', 'test_homog_errors', 'test_language_modulus',
           'test_mser', 'test_ondisk_find_patch_fpath_dominant_orientations',
           'test_score_normalization', 'test_show_gaussian_patches',
           'test_show_gaussian_patches2', 'testdata_annot_metadata',
           'testdata_binary_scores', 'testdata_blend',
           'testdata_dummy_matches', 'testdata_dummy_sift', 'testdata_hist',
           'testdata_imglist', 'testdata_matching_affine_inliers',
           'testdata_matching_affine_inliers_normalized',
           'testdata_nonmonotonic', 'testdata_patch', 'testdata_ratio_matches',
           'testdata_score_normalier', 'testdata_scores_labels',
           'testdata_sift2', 'testshow_extramargin_info',
           'to_undirected_edges', 'transform_around', 'transform_kpts',
           'transform_kpts_to_imgspace', 'transform_kpts_xys',
           'transform_points_with_homography', 'translation_mat3x3', 'trig',
           'try_svd', 'trytake', 'tune_flann', 'tune_flann2',
           'understanding_pseudomax_props', 'uniform_sample_hypersphere',
           'union_extents', 'unique_row_indexes', 'unique_rows',
           'unnormalize_transform', 'unstructure_rows',
           'unsupervised_multicut_labeling', 'util_math', 'verts_from_bbox',
           'verts_list_from_bboxes_list', 'warpAffine', 'warpHomog',
           'weighted_average_scoring', 'weighted_geometic_mean',
           'weighted_geometic_mean_unnormalized', 'whiten_xy_points',
           'wrap_histogram', 'wrapped_distance', 'zipcat', 'zipcompress',
           'zipcompress_safe', 'ziptake', 'zstar_value']



import numpy as np
import utool as ut
import ubelt as ub
import cv2
from vtool_ibeis import coverage_kpts


# TODO: integrate more
COVGRID_DEFAULT = ut.ParamInfoList('coverage_grid', [
    ut.ParamInfo('pxl_per_bin', 10, 'ppb', varyvals=[20, 5, 1]),
    ut.ParamInfo('grid_steps', 3, 'stps', varyvals=[1, 3, 7]),
    ut.ParamInfo('grid_sigma', 1.6, 'sigma', varyvals=[1.0, 1.6]),
])


def make_grid_coverage_mask(kpts, chipsize, weights, pxl_per_bin=4,
                            grid_steps=1, resize=False, out=None, grid_sigma=1.6):
    r"""
    Args:
        kpts (ndarray[float32_t, ndim=2]):  keypoint
        chipsize (tuple):  width, height
        weights (ndarray[float32_t, ndim=1]):
        pxl_per_bin (float):
        grid_steps (int):

    Returns:
        ndarray: weightgrid

    CommandLine:
        python -m vtool_ibeis.coverage_grid --test-make_grid_coverage_mask --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.coverage_grid import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> # build test data
        >>> kpts, chipsize, weights = coverage_kpts.testdata_coverage('tsukuba_l')
        >>> pxl_per_bin = 4
        >>> grid_steps = 2
        >>> # execute function
        >>> weightgrid = make_grid_coverage_mask(kpts, chipsize, weights, pxl_per_bin, grid_steps)
        >>> # verify result
        >>> result = str(weightgrid)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(weightgrid)
        >>> ut.show_if_requested()
    """
    import vtool_ibeis as vt
    coverage_gridtup = sparse_grid_coverage(
        kpts, chipsize, weights,
        pxl_per_bin=pxl_per_bin,
        grid_steps=grid_steps,
        grid_sigma=grid_sigma
    )
    gridshape = coverage_gridtup[0:2]
    neighbor_bin_weights, neighbor_bin_indices = coverage_gridtup[-2:]
    oldshape_indices = neighbor_bin_indices.shape
    newshape_indices = (np.prod(oldshape_indices[0:2]), oldshape_indices[2])
    neighbor_bin_indices =  neighbor_bin_indices.reshape(newshape_indices).T
    neighbor_bin_weights = neighbor_bin_weights.flatten()
    # Get flat indexing into gridbin
    neighbor_bin_flat_indices = np.ravel_multi_index(neighbor_bin_indices, gridshape)
    # Group by bins with weight
    unique_flatxs, grouped_flatxs = vt.group_indices(neighbor_bin_flat_indices)
    grouped_weights = vt.apply_grouping(neighbor_bin_weights, grouped_flatxs)
    # FIXME: boundary cases are not handled right because their vote is split
    # into the same bin and is fighting with itself durring the max
    max_weights = list(map(np.max, grouped_weights))
    if out is None:
        weightgrid = np.zeros(gridshape)
    else:
        # outvar specified
        weightgrid = out
        weightgrid[:] = 0
    unique_rows, unique_cols = np.unravel_index(unique_flatxs, gridshape)
    weightgrid[unique_rows, unique_cols] = max_weights
    #flat_weightgrid = np.zeros(np.prod(gridshape))
    #flat_weightgrid[unique_flatxs] = max_weight
    #ut.embed()
    #weightgrid = np.reshape(flat_weightgrid, gridshape)
    if resize:
        weightgrid = cv2.resize(weightgrid, chipsize,
                                interpolation=cv2.INTER_NEAREST)
    return weightgrid


def get_subbin_xy_neighbors(subbin_index00, grid_steps, num_cols, num_rows):
    """ Generate all neighbor of a bin
    subbin_index00 = left and up subbin index
    """
    subbin_index00 = np.floor(subbin_index00).astype(np.int32)
    subbin_x0, subbin_y0 = subbin_index00
    step_list = np.arange(1 - grid_steps, grid_steps + 1)
    offset_list = [
        # broadcast to the shape we will add too
        np.array([xoff, yoff])[:, None]
        for xoff, yoff in list(ut.iprod(step_list, step_list))]
    neighbor_subbin_index_list = [
        np.add(subbin_index00, offset)
        for offset in offset_list
    ]
    # Concatenate all subbin indexes into one array for faster vectorized op
    neighbor_bin_indices = np.dstack(neighbor_subbin_index_list).T

    # Clip with no wrapparound
    min_val = np.array([0, 0])
    max_val = np.array([num_cols - 1, num_rows - 1])

    np.clip(neighbor_bin_indices,
            min_val[None, None, :],
            max_val[None, None, :],
            out=neighbor_bin_indices)
    return neighbor_bin_indices


def compute_subbin_to_bins_dist(neighbor_bin_centers, subbin_xy_arr):
    _tmp = np.subtract(neighbor_bin_centers, subbin_xy_arr.T[None, :])
    neighbor_subbin_sqrddist_arr = np.power(_tmp, 2, out=_tmp).sum(axis=2)
    return neighbor_subbin_sqrddist_arr


def weighted_gaussian_falloff(neighbor_subbin_sqrddist_arr, weights, grid_sigma):
    import vtool_ibeis as vt
    _gaussweights = vt.gauss_func1d_unnormalized(neighbor_subbin_sqrddist_arr, grid_sigma)
    # If uncommented next line ensure each column sums to 1
    #np.divide(_gaussweights, _gaussweights.sum(axis=0)[None, :], out=_gaussweights)
    # Scale initial weights by the gaussian falloff
    neighbor_bin_weights = np.multiply(_gaussweights, weights[None, :])
    return neighbor_bin_weights


def sparse_grid_coverage(kpts, chipsize, weights, pxl_per_bin=.3, grid_steps=1, grid_sigma=1.6):
    r"""
    Args:
        kpts (ndarray[float32_t, ndim=2]):  keypoint
        chipsize (tuple):
        weights (ndarray):

    CommandLine:
        python -m vtool_ibeis.coverage_grid --test-sparse_grid_coverage --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.coverage_grid import *  # NOQA
        >>> kpts, chipsize, weights = coverage_kpts.testdata_coverage()
        >>> chipsize = (chipsize[0] + 50, chipsize[1])
        >>> pxl_per_bin = 3
        >>> grid_steps = 2
        >>> grid_sigma = 1.6
        >>> coverage_gridtup = sparse_grid_coverage(kpts, chipsize, weights, pxl_per_bin, grid_steps, grid_sigma)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> show_coverage_grid(*coverage_gridtup)
        >>> pt.show_if_requested()
    """
    import vtool_ibeis as vt
    # Compute grid size and stride
    chip_w, chip_h = chipsize
    # find enough rows to fit pxl_per_bin pixels into a grid dimension
    num_rows = max(vt.iround(chip_h / pxl_per_bin), 1)
    num_cols = max(vt.iround(chip_w / pxl_per_bin), 1)
    # stride is roughly equal in each direction, depending on rounding errors
    chipstride = np.array((chip_w / num_cols, chip_h / num_rows))
    # Find keypoint subbin locations relative to edge
    xy_arr = vt.get_xys(kpts)
    subbin_xy_arr = np.divide(xy_arr, chipstride[:, None])
    # Find subbin locations relative to center
    frac_subbin_index = np.subtract(subbin_xy_arr, .5)
    neighbor_bin_xy_indices = get_subbin_xy_neighbors(frac_subbin_index, grid_steps, num_cols, num_rows)
    # Find center
    neighbor_bin_centers = np.add(neighbor_bin_xy_indices, .5)
    # compute distance to neighbor
    neighbor_subbin_sqrddist_arr = compute_subbin_to_bins_dist(neighbor_bin_centers, subbin_xy_arr)
    # scale weights using guassia falloff
    neighbor_bin_weights = weighted_gaussian_falloff(neighbor_subbin_sqrddist_arr, weights, grid_sigma)
    # convert to rowcol
    neighbor_bin_indices = neighbor_bin_rc_indices = neighbor_bin_xy_indices[:, :, ::-1]  # NOQA

    coverage_gridtup = (
        num_rows, num_cols, subbin_xy_arr, neighbor_bin_centers,
        neighbor_bin_weights, neighbor_bin_indices)
    return coverage_gridtup


# VISUALIZATION FUNCS


def show_coverage_grid(num_rows, num_cols, subbin_xy_arr,
                       neighbor_bin_centers, neighbor_bin_weights,
                       neighbor_bin_indices, fnum=None, pnum=None):
    """
    visualizes the voting scheme on the grid. (not a mask, and no max)
    """
    import plottool_ibeis as pt
    import vtool_ibeis as vt
    import matplotlib as mpl

    if fnum is None:
        fnum = pt.next_fnum()
    fig = pt.figure(fnum, pnum=pnum)
    ax = fig.gca()
    x_edge_indices = np.arange(num_cols)
    y_edge_indices = np.arange(num_rows)
    x_center_indices = vt.hist_edges_to_centers(x_edge_indices)
    y_center_indices = vt.hist_edges_to_centers(y_edge_indices)
    x_center_grid, y_center_grid = np.meshgrid(x_center_indices, y_center_indices)
    ax.set_xticks(x_edge_indices)
    ax.set_yticks(y_edge_indices)
    # Plot keypoint loc
    ax.scatter(subbin_xy_arr[0], subbin_xy_arr[1], marker='o')
    # Plot Weighted Lines to Subbin
    pt_colors = pt.distinct_colors(len(subbin_xy_arr.T))
    segment_list = []
    color_list = []
    for subbin_centers, subbin_weights in zip(neighbor_bin_centers,
                                              neighbor_bin_weights):
        for pt_xys, center_xys, weight, color in zip(subbin_xy_arr.T, subbin_centers,
                                                     subbin_weights, pt_colors):
            # Adjsut weight to alpha for easier visualization
            alpha = weight
            INCRESE_ALPHA_VISIBILITY = True
            if INCRESE_ALPHA_VISIBILITY:
                min_viz_alpha = .05
                alpha = alpha * (1.0 - min_viz_alpha) + min_viz_alpha
                alpha **= 1.0
            #pt.plots.colorline(
            segment = np.vstack((pt_xys, center_xys))
            segment_list.append(segment)
            # Alpha becomes part of the colors
            color_list.append(list(color) + [alpha])
            # DO NOT USE PLOT VERY SLOW
            #ax.plot(*segment.T, color=color, alpha=alpha, lw=3)
    ax = pt.gca()
    # Plot all segments in single  line collection for speed
    # solid | dashed | dashdot | dotted
    lc = mpl.collections.LineCollection(segment_list, colors=color_list, linewidth=3, linestyles='solid')
    ax.add_collection(lc)
    # Plot Grid Center
    num_cells = num_cols * num_rows
    grid_alpha = min(.4, max(1 - (num_cells / 500), .1))
    grid_color = [.6, .6, .6, grid_alpha]
    #print(grid_color)
    # Plot grid cetner
    ax.scatter(x_center_grid, y_center_grid, marker='.', color=grid_color,
               s=grid_alpha)

    ax.set_xlim(0, num_cols - 1)
    ax.set_ylim(0, num_rows - 1)
    #-----
    pt.dark_background()
    ax.grid(True, color=[.3, .3, .3])
    ax.set_xticklabels([])
    ax.set_yticklabels([])


# TESTING FUNCS


def get_coverage_grid_gridsearch_configs():
    #varied_dict = {
    #    'pxl_per_bin': [.05, .3, 1.0],
    #    'grid_steps': [1, 3, 7],
    #    'grid_sigma': [1.0, 1.6],
    #}
    #slice_dict = {
    #    'pxl_per_bin' : slice(0, 3),
    #    'grid_steps'        : slice(0, 3),
    #    'grid_sigma'        : slice(0, 3),
    #}
    cfgdict_list, cfglbl_list = COVGRID_DEFAULT.get_gridsearch_input()
    # Make configuration for every parameter setting
    #cfgdict_list, cfglbl_list = ut.make_constrained_cfg_and_lbl_list(varied_dict, slice_dict=slice_dict)
    return cfgdict_list, cfglbl_list


def gridsearch_coverage_grid():
    """
    CommandLine:
        python -m vtool_ibeis.coverage_grid --test-gridsearch_coverage_grid --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.coverage_grid import *  # NOQA
        >>> import plottool_ibeis as pt
        >>> gridsearch_coverage_grid()
        >>> pt.show_if_requested()
    """
    import plottool_ibeis as pt
    fname = None  # 'easy1.png'
    kpts, chipsize, weights = coverage_kpts.testdata_coverage(fname)
    if len(kpts) > 100:
        kpts = kpts[::100]
        weights = weights[::100]
    cfgdict_list, cfglbl_list = get_coverage_grid_gridsearch_configs()
    coverage_gridtup_list = [
        sparse_grid_coverage(kpts, chipsize, weights, **cfgdict)
        for cfgdict in ub.ProgIter(cfgdict_list, desc='coverage grid')
    ]

    fnum = 1
    with ub.Timer('plotting gridsearch'):
        ut.interact_gridsearch_result_images(
            show_coverage_grid, cfgdict_list, cfglbl_list,
            coverage_gridtup_list, fnum=fnum, figtitle='coverage grid', unpack=True,
            max_plots=25)

    pt.iup()


def gridsearch_coverage_grid_mask():
    """
    CommandLine:
        python -m vtool_ibeis.coverage_grid --test-gridsearch_coverage_grid_mask --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.coverage_grid import *  # NOQA
        >>> import plottool_ibeis as pt
        >>> gridsearch_coverage_grid_mask()
        >>> pt.show_if_requested()
    """
    import plottool_ibeis as pt
    cfgdict_list, cfglbl_list = get_coverage_grid_gridsearch_configs()
    kpts, chipsize, weights = coverage_kpts.testdata_coverage('astro')
    gridmask_list = [
        255 *  make_grid_coverage_mask(kpts, chipsize, weights, **cfgdict)
        for cfgdict in ub.ProgIter(cfgdict_list, desc='coverage grid')
    ]
    NORMHACK = False
    if NORMHACK:
        gridmask_list = [
            255 * (gridmask / gridmask.max()) for gridmask in gridmask_list
        ]

    fnum = 1
    ut.interact_gridsearch_result_images(
        pt.imshow, cfgdict_list, cfglbl_list,
        gridmask_list, fnum=fnum, figtitle='coverage grid', unpack=False,
        max_plots=25)

    pt.iup()
    #pt.show_if_requested()


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.coverage_grid
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
Module for -- Confusion matrix, contingency, error matrix,

References:
    http://en.wikipedia.org/wiki/Confusion_matrix
"""
import utool as ut
import ubelt as ub
import numpy as np
import scipy.interpolate


def testdata_scores_labels():
    scores = [2,  3,  4,  6,  9,  9, 13, 17, 19, 22, 22, 23, 26, 26, 34, 59, 63, 75, 80, 81, 89]
    labels = [0,  0,  0,  0,  0,  1,  0,  0,  0,  1,  0,  0,  1,  0,  1,  1,  1,  1,  1,  1,  1]
    return scores, labels


def nan_to_num(arr, num):
    arr[np.isnan(arr)] = num
    return arr


class ConfusionMetrics(ub.NiceRepr):
    r"""
    Can compute average percision using the PASCAL definition

    References:
        http://www.flinders.edu.au/science_engineering/fms/School-CSEM/publications/tech_reps-research_artfcts/TRRA_2007.pdf
        http://www.alta.asn.au/events/altss_w2003_proc/altss/courses/powers/Bookmaker-all/200302-ICCS-Bookmaker.pdfcs
        http://www.cs.bris.ac.uk/Publications/Papers/1000704.pdf
        http://en.wikipedia.org/wiki/Information_retrieval
        http://en.wikipedia.org/wiki/Precision_and_recall
        https://en.wikipedia.org/wiki/Confusion_matrix
        http://scikit-learn.org/stable/modules/generated/sklearn.metrics.roc_curve.html#sklearn.metrics.roc_curve

    SeeAlso:
        sklearn.metrics.ranking._binary_clf_curve

    Notes:
        From oxford:
        Precision is defined as the ratio of retrieved positive images to the
          total number retrieved.
        Recall is defined as the ratio of the number of retrieved positive
          images to the total number of positive images in the corpus.

    Ignore:
        varname_list = 'tp, fp, fn, tn, fpr, tpr, tpa'.split(', ')
        lines = ['self.{varname} = {varname}'.format(varname=varname) for varname in varname_list]
        print(ut.indent('\n'.join(lines)))

    CommandLine:
        python -m vtool_ibeis.confusion ConfusionMetrics --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.confusion import *  # NOQA
        >>> scores, labels = testdata_scores_labels()
        >>> c = self = confusions = ConfusionMetrics().fit(scores, labels)
        >>> assert np.all(c.n_pos == c.n_tp + c.n_fn)
        >>> assert np.all(c.n_neg == c.n_tn + c.n_fp)
        >>> assert np.all(np.isclose(c.rp + c.rn, 1.0))
        >>> assert np.all(np.isclose(c.pp + c.pn, 1.0))
        >>> assert np.all(np.isclose(c.fpr, 1 - c.tnr))
        >>> assert np.all(np.isclose(c.fnr, 1 - c.tpr))
        >>> assert np.all(np.isclose(c.tpr, c.tp / c.rp))
        >>> assert np.all(np.isclose(c.tpa, c.tp / c.pp))
        >>> assert np.all(np.isclose(c.jacc, c.tp / (c.tp + c.fn + c.fp)))
        >>> assert np.all(np.isclose(c.mcc, np.sqrt(c.mk * c.bm)))
        >>> assert np.all(np.isclose(
        >>>     c.acc, (c.tpr + c.c * (1 - c.fpr)) / (1 + c.c)))
        >>> assert np.all(np.isclose(c.ppv, c.recall * c.prev / c.bias))
        >>> assert np.all(np.isclose(
        >>>    c.wracc, 4 * c.c * (c.tpr - c.fpr) / (1 + c.c) ** 2))
        >>> # xdoctest: +REQUIRES(--show)
        >>> confusions.draw_roc_curve()
        >>> ut.show_if_requested()
    """
    aliases = {
        'n_tp': {'n_hit', 'n_true_pos'},
        'n_tn': {'n_reject', 'n_true_neg'},
        'n_fp': {'n_false_alarm', 'n_false_pos'},
        'n_fn': {'n_miss', 'n_false_neg'},
        # -----
        'rp': {'real_pos', 'prev', 'prevalence'},
        'rn': {'real_neg'},
        'pp': {'pred_pos', 'bias'},
        'pn': {'pred_neg'},
        # -----
        'cs': {'class_odds', 'skew'},
        'cv': {'cost_ratio'},
        'cp': {'cost_pos'},
        'cn': {'cost_neg'},
        # -----
        'tp': {'true_pos', 'hit'},
        'tn': {'true_neg', 'reject'},
        'fp': {'false_pos', 'type1_error', 'false_alarm'},
        'fn': {'false_neg', 'type2_error', 'miss'},
        # -----
        'fpr': {'false_pos_rate', 'fallout'},
        'fnr': {'false_neg_rate', 'miss_rate'},
        'tpr': {'true_pos_rate', 'recall', 'sensitivity', 'hit_rate'},
        'tnr': {'true_neg_rate', 'inv_recall, specificity'},
        # -----
        'tpa': {'true_pos_acc', 'pos_predict_value', 'precision', 'ppv'},
        'tna': {'true_neg_acc', 'neg_predict_value', 'inv_precision', 'npv'},
        # -----
        'mk': {'markedness', 'deltaP', 'r_P'},
        'bm': {'informedness', 'bookmaker_informedness', 'deltaP\'', 'r_R'},
        # -----
        'mcc': {'matthews_correlation_coefficient'},
        'jacc': {'jaccard_coefficient'},
        'acc': {'accuracy', 'rand_accuracy', 'tea', 'ter'},
        'wracc': {'weighted_relative_accuracy'},
    }

    # the same things are called by lots of different names
    paper_alias = [
        ['dtp', 'determinant'],
        ['lr', 'liklihood-ratio'],
        ['nlr', 'negative-liklihood-ratio'],
        ['bmg', 'bookmarkG', 'bookmark_geometric_mean', 'mcc?'],
        ['evenness_R', 'PrevG2'],
        ['evenness_P', 'BiasG2'],
        ['rh', 'real_harmonic_mean'],
        ['ph', 'pred_harminic_mean'],
    ]

    # And they related to each other in interesting ways
    paper_relations = {
        'N'      : ['A + B + C + D'],
        'dtp'    : ['A * D - B * C'],
        'mk'     : ['dtp / (bias * (1 - bias))',
                    'dtp / biasG ** 2'],
        'bm'     : ['dtp / (prev * (1 - prev))'],
        'BiasG2' : ['bias * 1 - bias'],
        'lr'     : ['tpr / (1 - tnr)'],
        'nlr'    : ['tnr / (1 - tpr)'],
        'BMG'    : ['dtp / evenness_G'],
        'IBias'  : ['1 - Bias'],
        'etp'    : ['rp * pp', 'expected_true_positives'],
        'etn'    : ['rn * pn', 'expected_true_negatives'],
        'rh' : ['2 * rp * rn / (rp + rn)', 'real_harmonic_mean'],
        'ph' : ['2 * pp * pn / (pp + pn)', 'pred_harminic_mean'],
        'dp'     : ['tp - etp', 'dtp', '-dtn', '-(tn - etn)'],
        'deltap' : ['dtp - dtn', '2 * dp'],
        'kappa': ['deltap / (deltap + (fp + fn) / 2)']
    }

    # ROC Plot: tpr vs fpr
    # PN Plot: TP vs FP

    minimizing_metrics = {'fpr', 'fnr', 'fp', 'fn'}

    inv_aliases = {
        alias_key: std_key
        for std_key, alias_vals in aliases.items()
        for alias_key in set.union(alias_vals, {std_key})
    }

    def __init__(self):
        # Scalars
        self.n_pos = None
        self.n_neg = None
        self.n_samples = None

        # Threshold based
        self.thresholds = None
        self.n_tp = None
        self.n_fp = None
        self.n_fn = None
        self.n_tn = None

        # Can be set to weight the cost of errors
        self.cp = 1.0
        self.cn = 1.0

    # def __nice__(self):
    #     return '{}'.format(cfms.n_samples)

    @property
    def thresh(self):
        return self.thresholds

    # ----

    @property
    def cs(self):
        """ class ratio """
        return self.rn / self.rp

    @property
    def cv(self):
        """ ratio of cost of making a mistake"""
        return self.cn / self.cp

    @property
    def c(self):
        return self.cs * self.cv

    # -----

    @property
    def tp(self):
        """ true positive probability """
        return self.n_tp / self.n_samples

    @property
    def tn(self):
        """ true negative probability """
        return self.n_tn / self.n_samples

    @property
    def fp(self):
        """ false positive probability """
        return self.n_fp / self.n_samples

    @property
    def fn(self):
        """ false negative probability """
        return self.n_fn / self.n_samples

    # ----

    @property
    def rp(self):
        """ real positive probability """
        # return (self.tp + self.fn)
        return self.n_pos / self.n_samples

    @property
    def rn(self):
        """ real negative probability """
        # return (self.fp + self.tn)
        return self.n_neg / self.n_samples

    @property
    def pp(self):
        """ predicted positive probability """
        return (self.tp + self.fp)

    @property
    def pn(self):
        """ predicted negative probability """
        return (self.fn + self.tn)

    # ----

    @property
    def fpr(self):
        """ fallout, false positive rate """
        return self.n_fp / self.n_neg

    @property
    def fnr(self):
        """ miss rate, false negative rate """
        return self.n_fn / self.n_pos

    @property
    def tpr(self):
        """ sensitivity, recall, hit rate, tpr """
        return self.n_tp / self.n_pos

    @property
    def tnr(self):
        """ true negative rate, inverse recall """
        return self.n_tn / self.n_neg

    # ----

    @property
    def tpa(self):
        """ miss rate, false negative rate """
        with np.errstate(invalid='ignore'):
            return nan_to_num(self.n_tp / (self.n_tp + self.n_fp), 1.0)

    @property
    def tna(self):
        """ negative predictive value, inverse precision """
        with np.errstate(invalid='ignore'):
            return nan_to_num(self.n_tn / (self.n_tn + self.n_fn), 1.0)

    # ----

    @property
    def bm(self):
        """ bookmaker informedness """
        return self.tpr + self.tnr - 1

    @property
    def mk(self):
        """ markedness """
        return self.tpa + self.tna - 1

    # ---- other measures

    @property
    def auc_trap(self):
        # per threshold trapazoidal auc metric
        return (self.tpr + self.tnr) / 2

    @property
    def acc(self):
        """ accuracy """
        return self.tp + self.tn

    @property
    def sqrd_error(self):
        """ squared error """
        return np.sqrt(self.fpr ** 2 + self.fnr ** 2)

    @property
    def mcc(self):
        """ matthews correlation coefficient

        Also true that:
            mcc == np.sqrt(self.bm * self.mk)
        """
        mcc_numer = (self.tp * self.tn - self.fp * self.fn)
        mcc_denom = np.sqrt((self.tp + self.fp) * (self.tp + self.fn) *
                            (self.tn + self.fp) * (self.tn + self.fn))
        with np.errstate(invalid='ignore'):
            mcc = nan_to_num(mcc_numer / mcc_denom, 0.0)
        return mcc

    @property
    def jacc(self):
        """ jaccard coefficient """
        return self.n_tp / (self.n_samples - self.n_tn)
        # return self.tp / (self.tp + self.fn + self.fp)

    @property
    def wracc(self):
        """ weighted relative accuracy """
        return 4 * (self.recall - self.bias) * self.prev

    # --- alias names currently needed for compatability

    def __dir__(self):
        attrs = dir(object)
        attrs += list(self.__class__.__dict__.keys())
        attrs += list(self.__dict__.keys())
        attrs += self.inv_aliases.keys()
        attrs = sorted(set(attrs))
        return attrs

    def __getattr__(self, attr):
        try:
            std_attr = self.inv_aliases[attr]
        except KeyError:
            raise AttributeError(attr)
        return getattr(self, std_attr)

    # @property
    # def recall(self):
    #     return self.tpr

    # @property
    # def precision(self):
    #     return self.tpa

    # @property
    # def fallout(self):
    #     return self.fpr

    # --------------
    # Construtors
    # --------------

    def fit(self, scores, labels, verbose=False):
        scores = np.asarray(scores)
        labels = np.asarray(labels)
        # must be binary
        labels = labels.astype(bool)
        if verbose:
            print('[confusion] building confusion metrics.')
            print('[confusion]  * scores.shape=%r, scores.dtype=%r' %
                  (scores.shape, scores.dtype))
            print('[confusion]  * labels.shape=%r, labels.dtype=%r' %
                  (labels.shape, labels.dtype))

        # sklearn has much faster implementation
        # n_fp - count the number of false positives with score >= threshold[i]
        # n_tp - count the number of true positives with score >= threshold[i]
        try:
            from sklearn.metrics._ranking import _binary_clf_curve
        except ImportError:
            from sklearn.metrics.ranking import _binary_clf_curve

        n_fp, n_tp, thresholds = _binary_clf_curve(
            labels, scores, pos_label=1)

        n_samples = len(labels)
        n_pos = labels.sum()
        n_neg = n_samples - n_pos

        # Scalars
        self.n_samples = n_samples
        self.n_pos = n_pos
        self.n_neg = n_neg

        # Threshold based
        self.thresholds = thresholds
        self.n_tp = n_tp
        self.n_fp = n_fp
        self.n_fn = n_pos - n_tp
        self.n_tn = n_neg - n_fp
        return self

    @classmethod
    def from_tp_and_tn_scores(cls, tp_scores, tn_scores, verbose=False):
        scores = np.hstack([tp_scores, tn_scores])
        labels = np.array([True] * len(tp_scores) + [False] * len(tn_scores))
        self = cls().fit(scores, labels, verbose=verbose)
        return self

    # -------------------------------
    # Threshold-less Summary Measures
    # -------------------------------

    def get_ave_precision(self):
        precision = self.precision
        recall = self.recall
        recall_domain, p_interp = interpolate_precision_recall(precision, recall)
        return p_interp.sum() / p_interp.size

    @property
    def auc(self):
        """
        The AUC is a standard measure used to evaluate a binary classifier and
          represents the probability that a random correct case will
          receive a higher score than a random incorrect case.

        References:
            https://en.wikipedia.org/wiki/Receiver_operating_characteristic#Area_under_the_curve
        """
        # TODO: change name to represent it is a total measure
        import sklearn.metrics
        return sklearn.metrics.auc(self.fpr, self.tpr)

    # ---------------------
    # Threshold Choosers / Info
    # ---------------------

    def get_fpr_at_recall(self, target_recall):
        indicies = np.where(self.recall >= target_recall)[0]
        assert len(indicies) > 0, 'no recall at target level'
        func = scipy.interpolate.interp1d(self.recall, self.fpr)
        interp_fpr = func(target_recall)
        ## interpolate to target recall
        #right_index  = indicies[0]
        #right_recall = self.recall[right_index]
        #left_index   = right_index - 1
        #left_recall  = self.recall[left_index]
        #stepsize = right_recall - left_recall
        #alpha = (target_recall - left_recall) / stepsize
        #left_fpr   = self.fpr[left_index]
        #right_fpr  = self.fpr[right_index]
        #interp_fpp = (left_fpr * (1 - alpha)) + (right_fpr * (alpha))
        return interp_fpr

    def get_recall_at_fpr(self, target_fpr):
        indicies = np.where(self.fpr >= target_fpr)[0]
        assert len(indicies) > 0, 'no false positives at target level'
        func = scipy.interpolate.interp1d(self.fpr, self.tpr)
        interp_tpr = func(target_fpr)
        return interp_tpr

    def get_thresh_at_metric_max(self, metric):
        """
        metric = 'mcc'
        metric = 'fnr'
        """
        metric_values = getattr(self, metric)
        if False:
            idx = metric_values.argmax()
            thresh = self.thresholds[idx]
        else:
            # interpolated version
            import vtool_ibeis as vt
            thresh, max_value = vt.argsubmax(metric_values, self.thresholds)
        return thresh

    def get_thresh_at_metric(self, metric, value, maximize=None):
        r"""
        Gets a threshold for a binary classifier using a target metric and value

        Args:
            metric (str): name of metric like tpr or fpr
            value (float): corresponding numeric value

        Returns:
            float: thresh

        CommandLine:
            python -m vtool_ibeis.confusion get_thresh_at_metric
            python -m vtool_ibeis.confusion --exec-interact_roc_factory --show

        Example:
            >>> # xdoctest: +SKIP
            >>> from vtool_ibeis.confusion import *  # NOQA
            >>> scores, labels = testdata_scores_labels()
            >>> self = ConfusionMetrics().fit(scores, labels)
            >>> metric = 'tpr'
            >>> value = .85
            >>> thresh = self.get_thresh_at_metric(metric, value)
            >>> print('%s = %r' % (metric, value,))
            >>> result = ('thresh = %s' % (str(thresh),))
            >>> print(result)
            thresh = 22.5

        Ignore:
            metric = 'fpr'
            value = 1e-4
            self = cfms
            maximize = False

        interpolate_replbounds(metric_values, self.thresholds, 0, maximize=maximize)
        interpolate_replbounds(metric_values, self.thresholds, 1e-4, maximize=maximize)
        interpolate_replbounds(metric_values, self.thresholds, 1e-3, maximize=maximize)
        interpolate_replbounds(metric_values, self.thresholds, 1e-2, maximize=maximize)
        interpolate_replbounds(metric_values, self.thresholds, 1e-2, maximize=maximize)
        """
        if value == 'max':
            return self.get_thresh_at_metric_max(metric)
        # if value == 'min':
        #     return self.get_thresh_at_metric_min(metric)
        # TODO: Use interpoloation here and make tpr vs fpr a smooth funciton
        metric = self.inv_aliases[metric]
        metric_values = getattr(self, metric)
        if metric == 'fpr':
            # hack
            if len(metric_values) <= 1:
                return 1.0
        # maximize = metric not in self.minimizing_metrics
        if maximize is None:
            maximize = metric not in {'fpr'}
        thresh = interpolate_replbounds(metric_values, self.thresholds, value,
                                        maximize=maximize)
        return thresh

    def get_metric_at_metric(self, get_metric, at_metric, at_value,
                             subindex=False, tiebreaker='maxthresh'):
        """
        Finds the corresponding value of `get_metric` at a specific value of
        `at_metric`.


        get_metric = 'fpr'
        at_metric = 'tpr'
        at_value = .25
        self.rrr()

        self.get_metric_at_metric('fpr', 'tpr', .25)
        self.get_metric_at_metric('n_false_pos', 'tpr', .25)
        self.get_metric_at_metric('n_true_pos', 'tpr', .25)

        get_metric = 'n_true_pos'
        at_metric = 'n_false_pos'
        at_value = 0
        subindex = False
        """
        index = self.get_index_at_metric(at_metric, at_value,
                                         subindex=subindex,
                                         tiebreaker=tiebreaker)
        get_value = self.get_metric_at_index(get_metric, index)
        return get_value

    def get_index_at_metric(self, at_metric, at_value, subindex=False,
                            tiebreaker='maxthresh'):
        """
        Finds the index that is closet to the metric at a given value

        Args:
            tiebreaker (str): either 'minimize' or 'maximize'
                if 'maximize', then a larger threshold is considered better
                when resolving ambiguities. Otherwise a smaller thresh is
                better.

        Doctest:
            >>> from vtool_ibeis.confusion import *
            >>> pat1 = [0, 0, 0, 0]
            >>> pat2 = [0, 0, 1, 1]
            >>> pat3 = [0, 1, 1, 1]
            >>> pat4 = [1, 1, 1, 1]
            >>> pats = [pat1, pat2, pat3, pat4]
            >>> n = 4
            >>> import itertools as it
            >>> s = it.count(0)
            >>> # Create places of ambiguitiy and unambiguity
            >>> x = list(ub.flatten([[next(s)] * len(pat) for pat in pats for _ in range(n)]))
            >>> y = list(ub.flatten([pat for pat in pats for _ in range(n)]))
            >>> self = ConfusionMetrics().fit(x, y)
            >>> at_metric = 'n_false_pos'
            >>> at_value = 0
            >>> subindex = False
            >>> idx1 = self.get_index_at_metric(at_metric, at_value, subindex=False, tiebreaker='minthresh')
            >>> idx2 = self.get_index_at_metric(at_metric, at_value, subindex=False, tiebreaker= 'maxthresh')
            >>> assert idx1 == 3
            >>> assert idx2 == 0
        """
        import vtool_ibeis as vt
        at_arr = getattr(self, at_metric)

        if at_value in {'max', 'maximize'}:
            at_value = at_arr.max()
        elif at_value in {'min', 'minimize'}:
            at_value = at_arr.min()

        # Find point closest to the value
        distance = np.abs(at_arr - at_value)

        if subindex:
            # TODO: need to be able to figure out how to correctly break ties
            raise NotImplementedError('fixme use argsubminima2 and then other stuff')
            submin_x, submin_y = vt.argsubmin2(distance)
            return submin_x
        else:
            # idx = distance.argmin()
            idxs = np.where(distance == distance.min())[0]
            # If len(idxs) is bigger than 0 it is ambiguous
            if tiebreaker is None:
                return idxs
            else:
                if tiebreaker == 'maxthresh':
                    # If we want to maximize the thresh then take leftmost
                    idx = idxs[0]
                elif tiebreaker == 'minthresh':
                    # If we want to minimize the thresh then take rightmost
                    idx = idxs[-1]
                else:
                    raise KeyError('tiebreaker = {!r}'.format(tiebreaker))
            return idx

    def get_metric_at_index(self, metric, subindex):
        import vtool_ibeis as vt
        arr = getattr(self, metric)
        if isinstance(subindex, int):
            value = arr[subindex]
        else:
            value = vt.linear_interpolation(arr, subindex)
        return value

    def get_metric_at_thresh(self, metric, thresh):
        r"""
        Args:
            metric (str): name of a metric
            thresh (float): desired threshold

        Returns:
            float : value - metric value

        CommandLine:
            python -m vtool_ibeis.confusion --exec-get_metric_at_threshold

        Ignore:
            >>> self = cfms
            >>> metric = 'fpr'
            >>> thresh = 0

        Example:
            >>> # ENABLE_DOCTEST
            >>> from vtool_ibeis.confusion import *  # NOQA
            >>> scores, labels = testdata_scores_labels()
            >>> self = ConfusionMetrics().fit(scores, labels)
            >>> metric = 'tpr'
            >>> thresh = .8
            >>> thresh = [0, .1, .9, 1.0]
            >>> value = self.get_metric_at_thresh(metric, thresh)
            >>> result = ('(None, None) = %s' % (str((None, None)),))
            >>> print(result)
        """
        was_scalar = ut.isscalar(thresh)
        if was_scalar:
            thresh = [thresh]
        else:
            thresh = np.asarray(thresh)
        # Assert decreasing
        assert (len(self.thresholds) == 1 or
                self.thresholds[0] > self.thresholds[-1])
        sortx = np.argsort(self.thresholds)
        thresh_ = np.clip(thresh, self.thresholds[-1], self.thresholds[0])
        r = np.searchsorted(self.thresholds, thresh_, side='left', sorter=sortx)
        index_list = sortx[r]
        # index_list = [np.where(self.thresholds <= t)[0][0] for t in thresh]
        # sortx[r]
        # index_list = []
        # for t in thresh:
        #     try:
        #         index = np.nonzero(self.thresholds <= t)[0][0]
        #     except IndexError:
        #         print('warning: index error in get_metric_at_thresh t=%r' % (t,))
        #         index = len(self.thresholds) - 1
        #     index_list.append(index)
        # # value = self.__dict__[metric][index]
        value = [getattr(self, metric)[index] for index in index_list]
        if was_scalar:
            value = value[0]
        return value

    # --------------
    # Visualizations
    # --------------

    def draw_roc_curve(self, **kwargs):
        return draw_roc_curve(self.fpr, self.tpr, **kwargs)

    def draw_precision_recall_curve(self, nSamples=11, **kwargs):
        precision = self.precision
        recall = self.recall
        recall_domain, p_interp = interpolate_precision_recall(precision, recall, nSamples)
        return draw_precision_recall_curve(recall_domain, p_interp, **kwargs)

    def plot_vs(self, x_metric, y_metric):
        """
        x_metric = 'thresholds'
        y_metric = 'fpr'
        """
        import plottool_ibeis as pt
        # pt.qtensure()
        # xdata = self.thresholds
        xdata = getattr(self, x_metric)
        ydata_list = [getattr(self, y_metric)]
        pt.multi_plot(xdata, ydata_list,
                      # label_list=[y_metric],
                      xlabel=x_metric, marker='',
                      ylabel=y_metric, use_legend=True)

    def plot_metrics(self):
        import plottool_ibeis as pt
        metrics = [
            'mcc',
            'acc',
            'auc_trap'
            # 'tpa', 'tpr',
            # 'acc', 'sqrd_error',
            # 'auc_trap',
            # 'mk', 'bm'
        ]
        metrics = [
            'fnr',
            'fpr',
            'tpr',
            'tnr',
        ]
        xdata = self.thresholds
        ydata_list = [getattr(self, m) for m in metrics]
        pt.multi_plot(xdata, ydata_list, label_list=metrics,
                      xlabel='threshold', marker='',
                      ylabel='metric', use_legend=True)

    def show_mcc(self):
        import plottool_ibeis as pt
        pt.multi_plot(self.thresholds, [self.mcc], xlabel='threshold', marker='',
                      ylabel='MCC')
        pass


def interpolate_replbounds(xdata, ydata, pt, maximize=True):
    """
    xdata = np.array([.1, .2, .3, .4, .5])
    ydata = np.array([.1, .2, .3, .4, .5])
    pt = .35

    FIXME:
        if duplicate xdata is given bad things happen.
    BUG:
        in scipy.interpolate.interp1d
        If there is a duplicate xdata, then assume_sorted=False will
        sort ydata by xdata, but xdata should retain its initial ordering
        in places of ambuguity. Currently it does not.
    Args:
        xdata (ndarray):
        ydata (ndarray):
        pt (ndarray):

    Returns:
        float: interp_vals

    CommandLine:
        python -m vtool_ibeis.confusion --exec-interpolate_replbounds

    Example:
        >>> from vtool_ibeis.confusion import *  # NOQA
        >>> scores, labels = testdata_scores_labels()
        >>> self = ConfusionMetrics().fit(scores, labels)
        >>> xdata = self.tpr
        >>> ydata = self.thresholds
        >>> pt = 1.0
        >>> #xdata = self.fpr
        >>> #ydata = self.thresholds
        >>> #pt = 0.0
        >>> thresh = interpolate_replbounds(xdata, ydata, pt, maximize=True)
        >>> print('thresh = %r' % (thresh,))
        >>> thresh = interpolate_replbounds(xdata, ydata, pt, maximize=False)
        >>> print('thresh = %r' % (thresh,))

    Example:
        >>> # xdoctest: +SKIP
        >>> from vtool_ibeis.confusion import *  # NOQA
        >>> xdata = np.array([0.7,  0.8,  0.8,  0.9,  0.9, 0.9])
        >>> ydata = np.array([34,    26,   23,   22,   19,  17])
        >>> pt = np.array([.85, 1.0, -1.0])
        >>> interp_vals = interpolate_replbounds(xdata, ydata, pt)
        >>> result = ('interp_vals = %s' % (str(interp_vals),))
        >>> assert np.allclose(results, [ 22.5,  17.,   34., ])
        >>> print(result)
    """
    if not ut.issorted(xdata):
        if ut.issorted(xdata[::-1]):
            xdata = xdata[::-1]
            ydata = ydata[::-1]
        else:
            raise AssertionError('need to sort xdata and ydata in function')
            sortx = np.lexsort(np.vstack([np.arange(len(xdata)), xdata]))
            xdata = xdata.take(sortx, axis=0)
            ydata = ydata.take(sortx, axis=0)

    is_scalar = not ub.iterable(pt)
    #print('----')
    #print('xdata = %r' % (xdata,))
    #print('ydata = %r' % (ydata,))
    if is_scalar:
        pt = np.array([pt])
    minval = xdata.min()
    maxval = xdata.max()
    argx_min_list = np.argwhere(xdata == minval)
    argx_max_list = np.argwhere(xdata == maxval)
    argx_min = argx_min_list.min()
    argx_max = argx_max_list.max()
    lower_mask = pt < xdata[argx_min]
    upper_mask = pt > xdata[argx_max]
    interp_mask = ~np.logical_or(lower_mask, upper_mask)
    #if isinstance(pt, np.ndarray):
    dtype = np.result_type(np.float32, ydata.dtype)
    interp_vals = np.empty(pt.shape, dtype=dtype)
    interp_vals[lower_mask] = ydata[argx_min]
    interp_vals[upper_mask] = ydata[argx_max]

    # TODO: fix duplicate values depending on if higher or lower numbers are
    # desirable
    if True:
        # Grouping should be ok because xdata should be sorted
        # therefore groupxs are consecutive
        import vtool_ibeis as vt
        unique_vals, groupxs = vt.group_indices(xdata)
        grouped_ydata = vt.apply_grouping(ydata, groupxs)
        if maximize:
            sub_idxs = [idxs[np.argmax(ys)] for idxs, ys in zip(groupxs, grouped_ydata)]
        else:
            sub_idxs = [idxs[np.argmin(ys)] for idxs, ys in zip(groupxs, grouped_ydata)]
        sub_idxs = np.array(sub_idxs)
        xdata = xdata[sub_idxs]
        ydata = ydata[sub_idxs]

    if np.any(interp_mask):
        # FIXME: allow assume_sorted = False
        func = scipy.interpolate.interp1d(xdata, ydata, kind='linear', assume_sorted=True)
        interp_vals[interp_mask] = func(pt[interp_mask])
    if is_scalar:
        interp_vals = interp_vals[0]
    # interpolate to target recall
    #right_index  = indicies[0]
    #right_recall = self.recall[right_index]
    #left_index   = right_index - 1
    #left_recall  = self.recall[left_index]
    #stepsize = right_recall - left_recall
    #alpha = (target_recall - left_recall) / stepsize
    #left_fpr   = self.fpr[left_index]
    #right_fpr  = self.fpr[right_index]
    #interp_fpp = (left_fpr * (1 - alpha)) + (right_fpr * (alpha))
    return interp_vals


def interpolate_precision_recall(precision, recall, nSamples=11):
    """
    Interpolates precision as a function of recall p_{interp}(r)

    Reduce wiggles in average precision curve by taking interpolated values
    along a uniform sample.

    References:
        http://en.wikipedia.org/wiki/Information_retrieval#Average_precision
        http://en.wikipedia.org/wiki/Information_retrieval#Mean_Average_precision

    CommandLine:
        python -m vtool_ibeis.confusion --test-interpolate_precision_recall --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.confusion import *  # NOQA
        >>> scores, labels = testdata_scores_labels()
        >>> nSamples = 11
        >>> confusions = ConfusionMetrics().fit(scores, labels)
        >>> precision = confusions.precision
        >>> recall = confusions.recall
        >>> recall_domain, p_interp = interpolate_precision_recall(confusions.precision, recall, nSamples=11)
        >>> result = ub.repr2(p_interp, precision=1, with_dtype=True)
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> draw_precision_recall_curve(recall_domain, p_interp)
        >>> ut.show_if_requested()
        np.array([ 1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  1. ,  0.9,  0.9,  0.8,  0.6], dtype=np.float64)
    """
    if precision is None:
        return None, None

    recall_domain = np.linspace(0, 1, nSamples)
    if False:
        # normal interpolation
        func = scipy.interpolate.interp1d(recall, precision,
                                          bounds_error=False,
                                          fill_value=precision.max())
        p_interp = func(recall_domain)
    else:
        # Pascal interpolation
        #candidate_masks = recall >= recall_domain[:, None]
        #candidates_idxs_ = [np.where(mask)[0] for mask in candidate_masks]
        #chosen_idx = [-1 if len(idxs) == 0 else idxs.min() for idxs in  candidates_idxs_]
        #p_interp = precision[chosen_idx]
        def p_interp(r):
            precision_candidates = precision[recall >= r]
            if len(precision_candidates) == 0:
                return 0
            return precision_candidates.max()

        p_interp = np.array([p_interp(r) for r in recall_domain])
    return recall_domain, p_interp


def interact_roc_factory(confusions, target_tpr=None, show_operating_point=False):
    r"""
    Args:
        confusions (Confusions):

    CommandLine:
        python -m vtool_ibeis.confusion --exec-interact_roc_factory --show

    Example:
        >>> # xdoctest: +SKIP
        >>> from vtool_ibeis.confusion import *  # NOQA
        >>> scores, labels = testdata_scores_labels()
        >>> print('scores = %r' % (scores,))
        >>> confusions = ConfusionMetrics().fit(scores, labels)
        >>> print(ut.make_csv_table(
        >>>   [confusions.fpr, confusions.tpr, confusions.thresholds],
        >>>   ['fpr', 'tpr', 'thresh']))
        >>> # xdoctest: +REQUIRES(--show)
        >>> ROCInteraction = interact_roc_factory(confusions, target_tpr=.4, show_operating_point=True)
        >>> inter = ROCInteraction()
        >>> inter.show_page()
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> ut.show_if_requested()
    """
    from plottool_ibeis.abstract_interaction import AbstractInteraction

    class ROCInteraction(AbstractInteraction):
        """
        References:
            http://scipy-central.org/item/38/1/roc-curve-demo

        Notes:
            Sensitivity = true positive rate
            Specificity = true negative rate
        """
        def __init__(self, **kwargs):
            print('ROC Interact')
            super(ROCInteraction, self).__init__(**kwargs)
            self.confusions = confusions
            self.target_fpr = None
            self.show_operating_point = show_operating_point

        @staticmethod
        def static_plot(fnum, pnum, **kwargs):
            #print('ROC Interact2')
            kwargs['thresholds'] = kwargs.get('thresholds', confusions.thresholds)
            kwargs['show_operating_point'] = kwargs.get('show_operating_point',
                                                        show_operating_point)
            confusions.draw_roc_curve(fnum=fnum, pnum=pnum,
                                      target_tpr=target_tpr, **kwargs)

        def plot(self, fnum, pnum):
            #print('ROC Interact3')
            self.static_plot(fnum, pnum, target_fpr=self.target_fpr,
                             show_operating_point=self.show_operating_point)

        def on_click_inside(self, event, ex):
            self.target_fpr = event.xdata
            self.show_page()
            self.draw()

        def on_drag(self, event):
            # FIXME: blit
            if False:
                #print('Dragging ' + str(event.x) + ' ' + str(event.y))
                self.target_fpr = event.xdata
                self.show_page()
                #self.draw()
                if event.inaxes is not None:
                    self.fig.canvas.blit(event.inaxes.bbox)
                    #[blit(ax) event.canvas.figure.axes]

    return ROCInteraction


def draw_roc_curve(fpr, tpr, fnum=None, pnum=None, marker='', target_tpr=None,
                   target_fpr=None, thresholds=None, color=None, name=None,
                   label=None, show_operating_point=False):
    r"""
    Args:
        fpr (?):
        tpr (?):
        fnum (int):  figure number(default = None)
        pnum (tuple):  plot number(default = None)
        marker (str): (default = '-x')
        target_tpr (None): (default = None)
        target_fpr (None): (default = None)
        thresholds (None): (default = None)
        color (None): (default = None)
        show_operating_point (bool): (default = False)

    CommandLine:
        python -m vtool_ibeis.confusion --exec-draw_roc_curve --show --lightbg

    Example:
        >>> # xdoctest: +SKIP
        >>> from vtool_ibeis.confusion import *  # NOQA
        >>> scores, labels = testdata_scores_labels()
        >>> confusions = ConfusionMetrics().fit(scores, labels)
        >>> fpr = confusions.fpr
        >>> tpr = confusions.tpr
        >>> thresholds = confusions.thresholds
        >>> fnum = None
        >>> pnum = None
        >>> marker = 'x'
        >>> target_tpr = .85
        >>> target_fpr = None
        >>> color = None
        >>> show_operating_point = True
        >>> draw_roc_curve(fpr, tpr, fnum, pnum, marker, target_tpr, target_fpr,
        >>>   thresholds, color, show_operating_point)
        >>> ut.show_if_requested()
    """
    import plottool_ibeis as pt
    import sklearn.metrics
    if fnum is None:
        fnum = pt.next_fnum()

    # if color is None:
    #     color = (0.4, 1.0, 0.4) if pt.is_default_dark_bg() else  (0.1, 0.4, 0.4)

    roc_auc = sklearn.metrics.auc(fpr, tpr)

    title_suffix = ''

    if target_fpr is not None:
        #func = scipy.interpolate.interp1d(fpr, tpr, kind='linear', assume_sorted=False)
        #func = scipy.interpolate.interp1d(xdata, ydata, kind='nearest', assume_sorted=False)
        #interp_vals[interp_mask] = func(pt[interp_mask])
        target_fpr = np.clip(target_fpr, 0, 1)
        interp_tpr = interpolate_replbounds(fpr, tpr, target_fpr)
        choice_tpr = interp_tpr
        choice_fpr = target_fpr
    elif target_tpr is not None:
        target_tpr = np.clip(target_tpr, 0, 1)
        interp_fpr = interpolate_replbounds(tpr, fpr, target_tpr)
        choice_tpr = target_tpr
        choice_fpr = interp_fpr
    else:
        choice_tpr = None
        choice_fpr = None

    if choice_fpr is not None:
        choice_thresh = 0
        if thresholds is not None:
            try:
                index = np.nonzero(tpr >= choice_tpr)[0][0]
            except IndexError:
                index = len(thresholds) - 1
            choice_thresh = thresholds[index]
        #percent = ut.scalar_str(choice_tpr * 100).split('.')[0]
        #title_suffix = ', FPR%s=%05.2f%%' % (percent, choice_fpr)
        title_suffix = ''
        if show_operating_point:
            title_suffix = ', fpr=%.2f, tpr=%.2f, thresh=%.2f' % (
                choice_fpr, choice_tpr, choice_thresh)
    else:
        title_suffix = ''

    #if recall_domain is None:
    #    ave_p = np.nan
    #else:
    #    ave_p = p_interp.sum() / p_interp.size
    title = 'Receiver operating characteristic'
    if name and not label:
        title += ' (%s)' % (name,)
    if not label:
        title += '\n' + 'AUC=%.3f' % (roc_auc,)
    else:
        label += ' AUC=%.3f' % (roc_auc,)

    title += title_suffix

    label_list = None
    if label:
        label_list = [label]

    pt.multi_plot(fpr, [tpr], label_list=label_list, marker=marker,
                  color=color, fnum=fnum, pnum=pnum, title=title,
                  xlabel='False Positive Rate',
                  ylabel='True Positive Rate')

    # pt.plot2(fpr, tpr, marker=marker,
    #          x_label='False Positive Rate',
    #          y_label='True Positive Rate',
    #          unitbox=True, flipx=False, color=color, fnum=fnum, pnum=pnum,
    #          title=title)

    if False:
        # Interp does not work right because of duplicate values
        # in xdomain
        line_ = np.linspace(.11, .9, 20)
        #np.append([np.inf], np.diff(fpr)) > 0
        #np.append([np.inf], np.diff(tpr)) > 0
        unique_tpr_idxs = np.nonzero(np.append([np.inf], np.diff(tpr)) > 0)[0]
        unique_fpr_idxs = np.nonzero(np.append([np.inf], np.diff(fpr)) > 0)[0]

        pt.plt.plot(
            line_,
            interpolate_replbounds(fpr[unique_fpr_idxs], tpr[unique_fpr_idxs],
                                   line_), 'b-x')
        pt.plt.plot(
            interpolate_replbounds(tpr[unique_tpr_idxs], fpr[unique_tpr_idxs],
                                   line_), line_, 'r-x')
    if choice_fpr is not None:
        pt.plot(choice_fpr, choice_tpr, 'o', color=pt.PINK)


def draw_precision_recall_curve(recall_domain, p_interp, title_pref=None,
                                fnum=1, pnum=None,
                                color=None):
    import plottool_ibeis as pt
    if color is None:
        color = (0.4, 1.0, 0.4) if pt.is_default_dark_bg() else  (0.1, 0.4, 0.4)
    if recall_domain is None:
        recall_domain = np.array([])
        p_interp = np.array([])
    if recall_domain is None:
        ave_p = -1.0  # np.nan
    else:
        ave_p = p_interp.sum() / p_interp.size

    pt.plot2(recall_domain, p_interp, marker='o--',
              x_label='recall', y_label='precision', unitbox=True,
              flipx=False, color=color, fnum=fnum, pnum=pnum,
              title='Interplated Precision Vs Recall\n' + 'avep = %.3f'  % ave_p)
    #print('Interplated Precision')
    #print(ub.repr2(list(zip(recall_domain, p_interp))))
    #fig.show()


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.confusion
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



import numpy as np
import numpy.linalg as npl
from vtool_ibeis import linalg as ltool
from vtool_ibeis import image as gtool
import utool as ut
try:
    import cv2
except ImportError:
    print('ERROR: import cv2 is failing!')
    raise


def get_image_to_chip_transform(bbox, chipsz, theta):
    """
    transforms image space into chipspace

    Args:
        bbox   - bounding box of chip in image space
        chipsz - size of the chip
        theta  - rotation of the bounding box

    Sympy:
        # https://groups.google.com/forum/#!topic/sympy/k1HnZK_bNNA
        from vtool_ibeis.patch import *  # NOQA
        import sympy
        import sympy.abc
        theta = sympy.abc.theta

        x, y, w, h, target_area  = sympy.symbols('x y w h, a')
        gx, gy  = sympy.symbols('gx, gy')

        round = sympy.floor  # hack

        ht = sympy.sqrt(target_area * h / w)
        wt = w * ht / h
        cw_, ch_ = round(wt), round(ht)

        from vtool_ibeis import ltool
        T1 = ltool.translation_mat3x3(tx1, ty1, dtype=None)
        S  = ltool.scale_mat3x3(sx, sy, dtype=None)
        R  = ltool.rotation_mat3x3(-theta, sympy.sin, sympy.cos)
        T2 = ltool.translation_mat3x3(tx2, ty2, dtype=None)

        def add_matmul_hold_prop(mat):
            #import functools
            mat = sympy.Matrix(mat)
            def matmul_hold(other, hold=False):
                new = sympy.MatMul(mat, other, hold=hold)
                add_matmul_hold_prop(new)
                return new
            setattr(mat, 'matmul_hold', matmul_hold)
            return mat

        T1 = add_matmul_hold_prop(T1)
        T2 = add_matmul_hold_prop(T2)
        R = add_matmul_hold_prop(R)
        S = add_matmul_hold_prop(S)

        C = T2.multiply(R.multiply(S.multiply(T1)))
        sympy.simplify(C)

    """
    (x, y, w, h) = bbox
    (cw_, ch_) = chipsz
    tx1 = -(x + (w / 2.0))
    ty1 = -(y + (h / 2.0))
    sx = (cw_ / w)
    sy = (ch_ / h)
    tx2 = (cw_ / 2.0)
    ty2 = (ch_ / 2.0)
    # Translate from bbox center to (0, 0)
    T1 = ltool.translation_mat3x3(tx1, ty1)
    # Scale to chip height
    S  = ltool.scale_mat3x3(sx, sy)
    # Rotate to chip orientation
    R  = ltool.rotation_mat3x3(-theta)
    # Translate from (0, 0) to chip center
    T2 = ltool.translation_mat3x3(tx2, ty2)
    # Merge into single transformation (operate left-to-right aka data on left)
    C = T2.dot(R.dot(S.dot(T1)))
    return C


def _get_chip_to_image_transform(bbox, chipsz, theta):
    """ transforms chip space into imgspace
        bbox   - bounding box of chip in image space
        chipsz - size of the chip
        theta  - rotation of the bounding box
    """
    C    = get_image_to_chip_transform(bbox, chipsz, theta)
    invC = npl.inv(C)
    return invC


def extract_chip_from_gpath(gfpath, bbox, theta, new_size, interpolation=cv2.INTER_LANCZOS4):
    imgBGR = gtool.imread(gfpath)  # Read parent image
    chipBGR = extract_chip_from_img(imgBGR, bbox, theta, new_size, interpolation)
    return chipBGR


def extract_chip_into_square(imgBGR, bbox, theta, target_size):
    bbox_size = bbox[2:4]
    unpadded_dsize, ratio = gtool.resized_dims_and_ratio(bbox_size, target_size)
    chipBGR = extract_chip_from_img(imgBGR, bbox, theta, unpadded_dsize)
    chipBGR_square = gtool.embed_in_square_image(chipBGR, target_size)
    return chipBGR_square


def extract_chip_from_gpath_into_square(args):
    gfpath, bbox, theta, target_size = args
    imgBGR = gtool.imread(gfpath)  # Read parent image
    return extract_chip_into_square(imgBGR, bbox, theta, target_size)


def extract_chip_from_img(imgBGR, bbox, theta, new_size, interpolation=cv2.INTER_LANCZOS4):
    """ Crops chip from image ; Rotates and scales;

    ibs.show_annot_image(aid)[0].pt_save_and_view()

    Args:
        gfpath (str):
        bbox (tuple):  xywh
        theta (float):
        new_size (tuple): wy

    Returns:
        ndarray: chipBGR

    CommandLine:
        python -m vtool_ibeis.chip --test-extract_chip_from_img
        python -m vtool_ibeis.chip --test-extract_chip_from_img --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.chip import *  # NOQA
        >>> # build test data
        >>> imgBGR = gtool.imread(ut.grab_test_imgpath('carl.jpg'))
        >>> bbox = (100, 3, 100, 100)
        >>> theta = 0.0
        >>> new_size = (58, 34)
        >>> # execute function
        >>> chipBGR = extract_chip_from_img(imgBGR, bbox, theta, new_size)
        >>> # verify results
        >>> assert chipBGR.shape[0:2] == new_size[::-1], 'did not resize correctly'
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(chipBGR)
        >>> pt.show_if_requested()
    """
    # THE CULPRIT FOR MULTIPROCESSING FREEZES
    flags = interpolation
    #if True:
    M = get_image_to_chip_transform(bbox, new_size, theta)  # Build transformation
    chipBGR = cv2.warpAffine(imgBGR, M[0:2], tuple(new_size), flags=flags, borderMode=cv2.BORDER_CONSTANT)
    #else:
    #    # if theta == 0, not sure if this is better. Certainly not more general
    #    x, y, w, h = bbox
    #    roiBGR = imgBGR[y:y + h, x:x + w, :]
    #    chipBGR = cv2.resize(roiBGR, tuple(new_size), interpolation=interpolation)
    #chipBGR = gtool.warpAffine(imgBGR, M, new_size)  # Rotate and scale
    return chipBGR


def gridsearch_chipextract():
    r"""
    CommandLine:
        xdoctest -m ~/code/vtool_ibeis/vtool_ibeis/chip.py gridsearch_chipextract --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> # GRIDSEARCH
        >>> from vtool_ibeis.chip import *  # NOQA
        >>> gridsearch_chipextract()
        >>> ut.show_if_requested()
    """
    import cv2
    test_func = extract_chip_from_img
    if False:
        gpath = ut.grab_test_imgpath('carl.jpg')
        bbox = (100, 3, 100, 100)
        theta = 0.0
        new_size = (58, 34)
    else:
        gpath = '/media/raid/work/GZ_Master1/_ibsdb/images/1524525d-2131-8770-d27c-3a5f9922e9e9.jpg'
        bbox = (450, 373, 2062, 1124)
        theta = 0.0
        old_size = bbox[2:4]
        #target_area = 700 ** 2
        target_area = 1200 ** 2
        new_size = ScaleStrat.area(target_area, old_size)
        print('old_size = %r' % (old_size,))
        print('new_size = %r' % (new_size,))
        #new_size = (677, 369)
    imgBGR = gtool.imread(gpath)
    args = (imgBGR, bbox, theta, new_size)
    param_info = ut.ParamInfoList('extract_params', [
        ut.ParamInfo('interpolation', cv2.INTER_LANCZOS4,
                     varyvals=[
                         cv2.INTER_LANCZOS4,
                         cv2.INTER_CUBIC,
                         cv2.INTER_LINEAR,
                         cv2.INTER_NEAREST,
                         #cv2.INTER_AREA
                     ],)
    ])
    show_func = None
    # Generalize
    import plottool_ibeis as pt
    pt.imshow(imgBGR)  # HACK
    cfgdict_list, cfglbl_list = param_info.get_gridsearch_input(defaultslice=slice(0, 10))
    fnum = pt.ensure_fnum(None)
    if show_func is None:
        show_func = pt.imshow
    lbl = ut.get_funcname(test_func)
    cfgresult_list = [
        test_func(*args, **cfgdict)
        for cfgdict in ut.ProgressIter(cfgdict_list, lbl=lbl)
    ]
    onclick_func = None
    ut.interact_gridsearch_result_images(
        show_func, cfgdict_list, cfglbl_list,
        cfgresult_list, fnum=fnum,
        figtitle=lbl, unpack=False,
        max_plots=25, onclick_func=onclick_func)
    pt.iup()


class ScaleStrat(object):
    """
    Scaling strategies
    """

    @staticmethod
    def maxwh(target, orig_wh, tol=0):
        r"""
        The maximum dimension becomes target

        Args:
            target (int): target size

        Example:
            >>> # ENABLE_DOCTEST
            >>> import utool as ut
            >>> ut.assert_eq(ScaleStrat.maxwh(800, (190, 220)), (691, 800))
            >>> ut.assert_eq(ScaleStrat.maxwh(800, (220, 190)), (800, 691))
        """
        max_idx = np.argmax(orig_wh)
        orig_dim_size = orig_wh[max_idx]
        low, high = (target - tol, target + tol)
        if low <= orig_dim_size and orig_dim_size <= high:
            new_size = orig_wh
        else:
            scale_factor = target / orig_dim_size
            wt = int(round(orig_wh[0] * scale_factor))
            ht = int(round(orig_wh[1] * scale_factor))
            new_size = (wt, ht)
        return new_size

    @staticmethod
    def width(target, orig_wh, tol=0):
        r"""
        The width becomes target

        Args:
            target (int): target size

        Example:
            >>> # ENABLE_DOCTEST
            >>> import utool as ut
            >>> ut.assert_eq(ScaleStrat.width(800, (190, 220)), (800, 926))
            >>> ut.assert_eq(ScaleStrat.width(800, (220, 190)), (800, 691))
        """
        orig_dim_size = orig_wh[0]
        low, high = (target - tol, target + tol)
        if low <= orig_dim_size and orig_dim_size <= high:
            new_size = orig_wh
        else:
            scale_factor = target / orig_dim_size
            wt = int(round(orig_wh[0] * scale_factor))
            ht = int(round(orig_wh[1] * scale_factor))
            new_size = (wt, ht)
        return new_size

    @staticmethod
    def area(target, orig_wh, tol=0):
        r"""
        The area becomes target

        Args:
            target (int): target size

        Example:
            >>> # ENABLE_DOCTEST
            >>> import utool as ut
            >>> ut.assert_eq(ScaleStrat.area(800 ** 2, (190, 220)), (743, 861))
            >>> ut.assert_eq(ScaleStrat.area(800 ** 2, (220, 190)), (861, 743))
        """
        w, h = orig_wh
        area = w * h
        low, high = (target - tol, target + tol)
        if low <= area and area <= high:
            new_size = orig_wh
        else:
            ht = np.sqrt(target * h / w)
            wt = w * ht / h
            new_size = (int(round(wt)), int(round(ht)))
        return new_size


def get_scaled_size_with_dlen(target_dlen, w, h):
    r"""
    returns new_size which scales (w, h) as close to target_dlen as possible
    and maintains aspect ratio
    """
    #ht = np.sqrt(target_area * h / w)
    #wt = w * ht / h
    #new_size = (int(round(wt)), int(round(ht)))
    raise NotImplementedError()
    #return new_size


def compute_chip(gfpath, bbox, theta, new_size, filter_list=[],
                 interpolation=cv2.INTER_LANCZOS4):
    r""" Extracts a chip and applies filters

    DEPRICATE

    Args:
        gfpath (str):  image file path string
        bbox (tuple):  bounding box in the format (x, y, w, h)
        theta (float):  angle in radians
        new_size (tuple): must maintain the same aspect ratio or else you will get weirdness
        filter_list (list):

    Returns:
        ndarray: chipBGR -  cropped image

    CommandLine:
        python -m vtool_ibeis.chip --test-compute_chip --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.chip import *  # NOQA
        >>> from vtool_ibeis.util_math import TAU
        >>> # build test data
        >>> gfpath = ut.grab_test_imgpath('carl.jpg')
        >>> bbox = (100, 3, 100, 100)
        >>> theta = TAU / 8
        >>> new_size = (32, 32)
        >>> filter_list = []
        >>> # execute function
        >>> chipBGR = compute_chip(gfpath, bbox, theta, new_size, filter_list)
        >>> # verify results
        >>> assert chipBGR.shape[0:2] == new_size[::-1], 'did not resize correctly'
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> import vtool_ibeis as vt
        >>> pt.imshow(vt.draw_verts(vt.imread(gfpath), vt.scaled_verts_from_bbox(bbox, theta, 1, 1)), pnum=(1, 2, 1))
        >>> pt.imshow(chipBGR, pnum=(1, 2, 2))
        >>> pt.show_if_requested()
    """
    chipBGR = extract_chip_from_gpath(gfpath, bbox, theta, new_size, interpolation)
    chipBGR = apply_filter_funcs(chipBGR, filter_list)
    return chipBGR


def apply_filter_funcs(chipBGR, filter_funcs):
    """ applies a list of preprocessing filters to a chip

    DEPRICATE

    """
    chipBGR_ = chipBGR
    for func in filter_funcs:
        chipBGR_ = func(chipBGR)
    return chipBGR_


def get_extramargin_measures(bbox_gs, new_size, halfoffset_ms=(64, 64)):
    r"""
    Computes a detection chip with a bit of spatial context so the detection
    algorithm doesn't clip boundaries

    Returns:
        mbbox_gs, margin_size -
            margin bounding box in image size,
            size of entire margined chip,

    CommandLine:
        python -m vtool_ibeis.chip --test-get_extramargin_measures --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.chip import *  # NOQA
        >>> gfpath = ut.grab_test_imgpath('carl.jpg')
        >>> bbox_gs = [40, 40, 150, 150]
        >>> theta = .15 * (np.pi * 2)
        >>> new_size = (150, 150)
        >>> halfoffset_ms = (32, 32)
        >>> mbbox_gs, margin_size = get_extramargin_measures(bbox_gs, new_size, halfoffset_ms)
        >>> # xdoctest: +REQUIRES(--show)
        >>> testshow_extramargin_info(gfpath, bbox_gs, theta, new_size, halfoffset_ms, mbbox_gs, margin_size)
    """
    # _ex denotes an expanded version
    # There are three spaces we are working in here
    # chip _cs (the space of the original chip)
    # margin _ms (the margin chip has the scale of chip space with padding)
    # imagespace _gs (the space using in bbox_gs specification)
    x_gs, y_gs, w_gs, h_gs = bbox_gs
    if w_gs == 0 or  h_gs == 0:
        raise ValueError('Bounding box has no area')
    w_cs, h_cs = new_size
    # Extra margin in chip space
    xo_ms, yo_ms = halfoffset_ms
    # Compute size of margin chip
    mw, mh = (w_cs + (2 * xo_ms), h_cs + (2 * yo_ms))
    margin_size = (mw, mh)
    # Get the conversion from chip to image space
    sx, sy = (w_gs / w_cs, h_gs / h_cs)
    # Convert the chip offsets to image space
    halfoffset_gs = ((sx * xo_ms), (sy * yo_ms))
    xo_gs, yo_gs = halfoffset_gs
    # Find the size of the expanded margin bbox in image space
    mbbox_gs = (x_gs - xo_gs, y_gs - yo_gs,
                w_gs + (2 * xo_gs), h_gs + (2 * yo_gs))
    return mbbox_gs, margin_size


def testshow_extramargin_info(gfpath, bbox_gs, theta, new_size, halfoffset_ms, mbbox_gs, margin_size):
    import plottool_ibeis as pt
    import vtool_ibeis as vt

    imgBGR = vt.imread(gfpath)
    chipBGR = compute_chip(gfpath, bbox_gs, theta, new_size, [])
    mchipBGR = compute_chip(gfpath, mbbox_gs, theta, margin_size, [])

    #index = 0
    w_cs, h_cs = new_size
    xo_ms, yo_ms = halfoffset_ms
    bbox_ms = [xo_ms, yo_ms, w_cs, h_cs]

    verts_gs = vt.scaled_verts_from_bbox(bbox_gs, theta, 1, 1)
    expanded_verts_gs = vt.scaled_verts_from_bbox(mbbox_gs, theta, 1, 1)
    expanded_verts_ms = vt.scaled_verts_from_bbox(bbox_ms, 0, 1, 1)
    # topheavy
    imgBGR = vt.draw_verts(imgBGR, verts_gs)
    imgBGR = vt.draw_verts(imgBGR, expanded_verts_gs)

    mchipBGR = vt.draw_verts(mchipBGR, expanded_verts_ms)

    fnum = 1
    pt.imshow(imgBGR, pnum=(1, 3, 1), fnum=fnum, title='original image')
    pt.gca().set_xlabel(str(imgBGR.shape))
    pt.imshow(chipBGR, pnum=(1, 3, 2), fnum=fnum, title='original chip')
    pt.gca().set_xlabel(str(chipBGR.shape))
    pt.imshow(mchipBGR, pnum=(1, 3, 3), fnum=fnum,
              title='scaled chip with expanded margin.\n(orig margin drawn in orange)')
    pt.gca().set_xlabel(str(mchipBGR.shape))

    pt.show_if_requested()
    #pt.imshow(chipBGR)


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.chip
    """
    import xdoctest
    xdoctest.doctest_module(__file__)




"""
References:
    % Single-image noise level estimation for blind denoising.
    % http://www.ok.ctrl.titech.ac.jp/res/NLE/TIP2013-noise-level-estimation06607209.pdfhttp://www.ok.ctrl.titech.ac.jp/res/NLE/TIP2013-noise-level-estimation06607209.pdf
"""
import utool as ut
import ubelt as ub
import numpy as np
import cv2


def compute_average_contrast(img):
    """
    CommandLine:
        python -m vtool_ibeis.quality_classifier --exec-compute_average_contrast --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.quality_classifier import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> img_fpath = ut.grab_test_imgpath('carl')
        >>> img = vt.imread(img_fpath, grayscale=True)
        >>> average_contrast, gradmag_sqrd = compute_average_contrast(img)
        >>> # xdoctest: +REQUIRES(module:plottool_ibeis)
        >>> import plottool_ibeis as pt
        >>> # xdoctest: +REQUIRES(--show)
        >>> pt.figure(fnum=1)
        >>> pt.plt.imshow(gradmag_sqrd)
        >>> ut.show_if_requested()
    """
    ksize = 5
    assert img.dtype == np.uint8
    img_ = img.astype(np.float64) / 255.0
    gradx = cv2.Sobel(img_, cv2.CV_64F, 1, 0, ksize=ksize)
    grady = cv2.Sobel(img_, cv2.CV_64F, 0, 1, ksize=ksize)
    gradmag_sqrd = (gradx ** 2) + (grady ** 2)
    total_contrast = gradmag_sqrd.sum()

    average_contrast = total_contrast / np.prod(gradmag_sqrd.shape)
    return average_contrast, gradmag_sqrd


def contrast_measures(img):
    ksize = 5
    assert img.dtype == np.uint8
    img_ = img.astype(np.float64) / 255.0
    gradx = cv2.Sobel(img_, cv2.CV_64F, 1, 0, ksize=ksize)
    grady = cv2.Sobel(img_, cv2.CV_64F, 0, 1, ksize=ksize)
    gradmag_sqrd = (gradx ** 2) + (grady ** 2)
    total_contrast = gradmag_sqrd.sum()

    average_contrast = total_contrast / np.prod(gradmag_sqrd.shape)

    std_dev = img.std()

    bin_count, bin_lbls = np.histogram(img)
    entropy = -np.sum(bin_count * np.log2(bin_count))

    import skimage.morphology
    import skimage.filters.rank
    entropyimg = skimage.filters.rank.entropy(img, skimage.morphology.disk(2))
    ave_local_entropy = entropyimg.sum() / np.prod(entropyimg.shape)

    contrast_measures = [average_contrast, std_dev, entropy, ave_local_entropy]
    return contrast_measures


def test_average_contrast():
    import vtool_ibeis as vt
    ut.get_valid_test_imgkeys()
    img_fpath_list = [ut.grab_test_imgpath(key) for key in ut.get_valid_test_imgkeys()]
    img_list = [vt.imread(img, grayscale=True) for img in img_fpath_list]
    avecontrast_list = np.array([compute_average_contrast(img) for img in img_list])
    import plottool_ibeis as pt
    nCols = len(img_list)
    fnum = None
    if fnum is None:
        fnum = pt.next_fnum()
    pt.figure(fnum=fnum, pnum=(2, 1, 1))
    sortx = avecontrast_list.argsort()
    y_list = avecontrast_list[sortx]
    x_list = np.arange(0, nCols) + .5
    pt.plot(x_list, y_list, 'bo-')
    sorted_imgs = list(ub.take(img_list, sortx))

    for px, img in ub.ProgIter(enumerate(sorted_imgs, start=1)):
        pt.imshow(img, fnum=fnum, pnum=(2, nCols, nCols + px))


def fourier_devtest(img):
    r"""
    Args:
        img (ndarray[uint8_t, ndim=2]):  image data

    CommandLine:
        python -m vtool_ibeis.quality_classifier --test-fourier_devtest --show

    References:
        http://opencv-python-tutroals.readthedocs.org/en/latest/py_tutorials/py_imgproc/py_transforms/py_fourier_transform/py_fourier_transform.html
        http://cns-alumni.bu.edu/~slehar/fourier/fourier.html

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.quality_classifier import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> img_fpath = ut.grab_test_imgpath('carl')
        >>> img = vt.imread(img_fpath, grayscale=True)
        >>> magnitude_spectrum = fourier_devtest(img)
    """
    import plottool_ibeis as pt
    def pad_img(img):
        rows, cols = img.shape
        nrows = cv2.getOptimalDFTSize(rows)
        ncols = cv2.getOptimalDFTSize(cols)
        right = ncols - cols
        bottom = nrows - rows
        bordertype = cv2.BORDER_CONSTANT
        nimg = cv2.copyMakeBorder(img, 0, bottom, 0, right, bordertype, value=0)
        return nimg

    def convert_to_fdomain(img):
        dft = cv2.dft(img.astype(np.float32), flags=cv2.DFT_COMPLEX_OUTPUT)
        #dft_shift = np.fft.fftshift(dft)
        return dft

    def convert_from_fdomain(dft):
        img = cv2.idft(dft)
        img = cv2.magnitude(img[:, :, 0], img[:, :, 1])
        img /= img.max()
        return img * 255.0

    def get_fdomain_mag(dft_shift):
        magnitude_spectrum = np.log(cv2.magnitude(dft_shift[:, :, 0], dft_shift[:, :, 1]))
        return magnitude_spectrum

    def imgstats(img):
        print('stats:')
        print('    dtype = %r ' % (img.dtype,))
        print('    ' + ut.get_stats_str(img, axis=None))

    nimg = pad_img(img)
    dft = convert_to_fdomain(nimg)
    #freq_domain = np.fft.fft2(img)
    #freq_domain_shift = np.fft.fftshift(freq_domain)

    rows, cols = nimg.shape
    crow, ccol = rows / 2 , cols / 2
    # create a mask first, center square is 1, remaining all zeros
    mask = np.zeros((rows, cols, 2), np.uint8)
    mask[crow - 30:crow + 30, ccol - 30:ccol + 30] = 1

    dft_mask = np.fft.ifftshift(np.fft.fftshift(dft) * mask)
    img_back = convert_from_fdomain(dft_mask)

    imgstats(dft)
    imgstats(mask)
    imgstats(nimg)
    imgstats(nimg)

    print('nimg.shape = %r' % (nimg.shape,))
    print('dft_shift.shape = %r' % (dft.shape,))

    if ut.show_was_requested():
        #import plottool_ibeis as pt
        next_pnum = pt.make_pnum_nextgen(nRows=3, nCols=2)
        pt.imshow(nimg, pnum=next_pnum(), title='nimg')
        pt.imshow(20 * get_fdomain_mag(dft), pnum=next_pnum(), title='mag(f)')
        pt.imshow(20 * get_fdomain_mag(dft_mask), pnum=next_pnum(), title='dft_mask')
        pt.imshow(img_back, pnum=next_pnum(), title='img_back')
        pt.show_if_requested()


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.quality_classifier
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



from __future__ import absolute_import, division, print_function
from six.moves import range
import vtool_ibeis.keypoint as ktool
import vtool_ibeis.linalg as ltool
import numpy as np
import utool as ut
from vtool_ibeis import demodata
from vtool_ibeis.util_math import TAU


def test_get_invR_mats_orientation():
    theta1 = TAU / 8
    theta2 = -TAU / 8
    theta3 = 0
    theta4 = 7 * TAU / 8

    invV_mats = demodata.get_dummy_invV_mats()

    def R_mats(theta):
        return np.array([ltool.rotation_mat2x2(theta) for _ in range(len(invV_mats))])

    def test_rots(theta):
        invVR_mats = ltool.matrix_multiply(invV_mats, R_mats(theta))
        _oris = ktool.get_invVR_mats_oris(invVR_mats)
        print('________')
        print('theta = %r' % (theta % TAU,))
        print('b / a = %r' % (_oris,))
        passed, error = ut.almost_eq(_oris, theta % TAU, ret_error=True)
        try:
            assert np.all(passed)
        except AssertionError as ex:
            ut.printex(ex, 'rotation unequal', key_list=['passed',
                                                            'error'])

    test_rots(theta1)
    test_rots(theta2)
    test_rots(theta3)
    test_rots(theta4)


if __name__ == '__main__':
    test_get_invR_mats_orientation()



import numpy as np
import utool as ut
import ubelt as ub


def testdata_blend(scale=128):
    import vtool_ibeis as vt
    img_fpath = ut.grab_test_imgpath('astro')
    img1 = vt.imread(img_fpath)
    rng = np.random.RandomState(0)
    img2 = vt.perlin_noise(img1.shape[0:2], scale=scale, rng=rng)[None, :].T
    img1 = vt.rectify_to_float01(img1)
    img2 = vt.rectify_to_float01(img2)
    return img1, img2


def gridsearch_image_function(param_info, test_func, args=tuple(), show_func=None):
    """
    gridsearch for a function that produces a single image
    """
    import plottool_ibeis as pt
    cfgdict_list, cfglbl_list = param_info.get_gridsearch_input(defaultslice=slice(0, 10))
    fnum = pt.ensure_fnum(None)
    if show_func is None:
        show_func = pt.imshow
    lbl = ut.get_funcname(test_func)
    cfgresult_list = [
        test_func(*args, **cfgdict)
        for cfgdict in ub.ProgIter(cfgdict_list, desc=lbl)
    ]
    onclick_func = None
    ut.interact_gridsearch_result_images(
        show_func, cfgdict_list, cfglbl_list,
        cfgresult_list, fnum=fnum,
        figtitle=lbl, unpack=False,
        max_plots=25, onclick_func=onclick_func)
    pt.iup()


def ensure_alpha_channel(img, alpha=1.0):
    import vtool_ibeis as vt
    img = vt.rectify_to_float01(img)
    c = vt.get_num_channels(img)
    if c == 4:
        return img
    else:
        alpha_channel = np.full(img.shape[0:2], fill_value=alpha, dtype=img.dtype)
        if c == 3:
            return np.dstack([img, alpha_channel])
        elif c == 1:
            return np.dstack([img, img, img, alpha_channel])
        else:
            raise ValueError('unknown dim')


def ensure_grayscale(img, colorspace_hint='BGR'):
    import vtool_ibeis as vt
    img = vt.rectify_to_float01(img)
    c = vt.get_num_channels(img)
    if c == 1:
        return img
    else:
        return vt.convert_colorspace(img, 'gray', colorspace_hint)


def overlay_alpha_images(img1, img2):
    """
    places img1 on top of img2 respecting alpha channels

    References:
        http://stackoverflow.com/questions/25182421/overlay-numpy-alpha
    """
    import vtool_ibeis as vt
    img1 = vt.rectify_to_float01(img1)
    img2 = vt.rectify_to_float01(img2)

    img1, img2 = vt.make_channels_comparable(img1, img2)

    # print('img1.shape = {!r}'.format(img1.shape))
    # print('img1.dtype = {!r}'.format(img1.dtype))
    # print('img1.max() = {!r}'.format(img1.max()))

    # print('img2.shape = {!r}'.format(img2.shape))
    # print('img2.dtype = {!r}'.format(img2.dtype))
    # print('img2.max() = {!r}'.format(img2.max()))

    c1 = vt.get_num_channels(img1)
    c2 = vt.get_num_channels(img2)
    if c1 == 4:
        alpha1 = img1[:, :, 3]
    else:
        alpha1 = np.ones(img1.shape[0:2], dtype=img1.dtype)

    if c2 == 4:
        alpha2 = img2[:, :, 3]
    else:
        alpha2 = np.ones(img2.shape[0:2], dtype=img2.dtype)

    rgb1 = img1[:, :, 0:3]
    rgb2 = img2[:, :, 0:3]

    alpha3 = alpha1 + alpha2 * (1 - alpha1)
    rgb3 = rgb1 * alpha1[..., None] + rgb2 * alpha2[..., None]

    numer1 = (rgb1 * alpha1[..., None])
    numer2 = (rgb2 * alpha2[..., None] * (1.0 - alpha1[..., None]))
    rgb3 = (numer1 + numer2) / alpha3[..., None]

    # img3 = np.dstack([rgb3, alpha3[..., None]])
    return rgb3


def blend_images(img1, img2, mode='average', **kwargs):
    """
    Args:
        img1 (np.ndarray): first image
        img2 (np.ndarray): second image
        mode (str): can be average, multiply, or overlay
    """
    if mode == 'average':
        return blend_images_average(img1, img2, **kwargs)
    elif mode == 'multiply':
        return blend_images_multiply(img1, img2, **kwargs)
    elif mode == 'overlay':
        return overlay_alpha_images(img1, img2)
    else:
        raise ValueError('mode = %r' % (mode,))


def blend_images_average(img1, img2, alpha=.5):
    r"""
    Args:
        img1 (ndarray[uint8_t, ndim=2]):  image data
        img2 (ndarray[uint8_t, ndim=2]):  image data
        alpha (float): (default = 0.5)

    Returns:
        ndarray: imgB

    References:
        https://en.wikipedia.org/wiki/Blend_modes

    CommandLine:
        python -m vtool_ibeis.blend blend_images_average:0 --show
        python -m vtool_ibeis.blend blend_images_average:1 --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.blend import *  # NOQA
        >>> alpha = 0.8
        >>> img1, img2 = testdata_blend()
        >>> imgB = blend_images_average(img1, img2, alpha)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(imgB)
        >>> ut.show_if_requested()

    Ignore:
        >>> # GRIDSEARCH
        >>> from vtool_ibeis.blend import *  # NOQA
        >>> test_func = blend_images_average
        >>> args = testdata_blend()
        >>> param_info = ut.ParamInfoList('blend_params', [
        ...    ut.ParamInfo('alpha', .8, 'alpha=',
        ...                 varyvals=np.linspace(0, 1.0, 25).tolist()),
        ... ])
        >>> gridsearch_image_function(param_info, test_func, args)
        >>> ut.show_if_requested()
    """
    #assert img1.shape == img2.shape, 'chips must be same shape to blend'
    #imgB = np.zeros(img2.shape, dtype=img2.dtype)
    #assert img1.min() >= 0 and img1.max() <= 1
    #assert img2.min() >= 0 and img2.max() <= 1
    if isinstance(alpha, np.ndarray):
        import vtool_ibeis as vt
        img1, img2 = vt.make_channels_comparable(img1, img2)
        img1, alpha = vt.make_channels_comparable(img1, alpha)
        img2, alpha = vt.make_channels_comparable(img2, alpha)
        imgB = (img1 * (1.0 - alpha)) + (img2 * (alpha))
    else:
        imgB = (img1 * (1.0 - alpha)) + (img2 * (alpha))
    #assert imgB.min() >= 0 and imgB.max() <= 1
    return imgB


def blend_images_average_stack(images, alpha=None):
    r"""
    Args:
        img1 (ndarray[uint8_t, ndim=2]):  image data
        img2 (ndarray[uint8_t, ndim=2]):  image data
        alpha (float): (default = 0.5)

    Returns:
        ndarray: imgB

    References:
        https://en.wikipedia.org/wiki/Blend_modes

    CommandLine:
        python -m vtool_ibeis.blend --test-blend_images_average:0 --show
        python -m vtool_ibeis.blend --test-blend_images_average:1 --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.blend import *  # NOQA
        >>> alpha = 0.8
        >>> img1, img2 = testdata_blend()
        >>> imgB = blend_images_average(img1, img2, alpha)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(imgB)
        >>> ut.show_if_requested()
    """
    if alpha is None:
        alpha = [1 / len(images)] * len(images)

    assert np.isclose(sum(alpha), 1.0)

    # Make all images comparable
    imgT = images[0]
    import vtool_ibeis as vt
    for img in images[1:]:
        imgT, _ = vt.make_channels_comparable(imgT, img)
    images = [vt.make_channels_comparable(img, imgT)[0] for img in images]
    imgB = np.sum([img * a for img, a in zip(images, alpha)], axis=0)
    #assert imgB.min() >= 0 and imgB.max() <= 1
    return imgB


def blend_images_mult_average(img1, img2, alpha=.5):
    r"""
    Args:
        img1 (ndarray[uint8_t, ndim=2]):  image data
        img2 (ndarray[uint8_t, ndim=2]):  image data
        alpha (float): (default = 0.5)

    Returns:
        ndarray: imgB

    References:
        https://en.wikipedia.org/wiki/Blend_modes

    CommandLine:
        python -m vtool_ibeis.blend --test-blend_images_mult_average:0 --show
        python -m vtool_ibeis.blend --test-blend_images_mult_average:1 --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.blend import *  # NOQA
        >>> alpha = 0.8
        >>> img1, img2 = testdata_blend()
        >>> imgB = blend_images_mult_average(img1, img2, alpha)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(imgB)
        >>> ut.show_if_requested()

    Ignore:
        >>> # GRIDSEARCH
        >>> from vtool_ibeis.blend import *  # NOQA
        >>> test_func = blend_images_mult_average
        >>> args = testdata_blend()
        >>> param_info = ut.ParamInfoList('blend_params', [
        ...    ut.ParamInfo('alpha', .8, 'alpha=',
        ...                 varyvals=np.linspace(0, 1.0, 9).tolist()),
        ... ])
        >>> gridsearch_image_function(param_info, test_func, args)
        >>> ut.show_if_requested()
    """
    #assert img1.shape == img2.shape, 'chips must be same shape to blend'
    #imgB = np.zeros(img2.shape, dtype=img2.dtype)
    #assert img1.min() >= 0 and img1.max() <= 1
    #assert img2.min() >= 0 and img2.max() <= 1
    import vtool_ibeis as vt
    img1_ = vt.rectify_to_float01(img1)
    img2_ = vt.rectify_to_float01(img2)
    img1_, img2_ = vt.make_channels_comparable(img1_, img2_)

    mult_ave = blend_images_multiply(img1_, img2_, .5)
    if alpha < .5:
        imgB = blend_images_average(img1_, mult_ave, alpha * 2)
    else:
        imgB = blend_images_average(mult_ave, img2_, (alpha - .5) * 2)
    #assert imgB.min() >= 0 and imgB.max() <= 1
    return imgB


def blend_images_multiply(img1, img2, alpha=0.5):
    r"""
    Args:
        img1 (ndarray[uint8_t, ndim=2]):  image data
        img2 (ndarray[uint8_t, ndim=2]):  image data
        alpha (float): (default = 0.5)

    Returns:
        ndarray: imgB


    References:
        https://en.wikipedia.org/wiki/Blend_modes

    CommandLine:
        python -m vtool_ibeis.blend --test-blend_images_multiply:0 --show
        python -m vtool_ibeis.blend --test-blend_images_multiply:1 --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from vtool_ibeis.blend import *  # NOQA
        >>> alpha = 0.8
        >>> img1, img2 = testdata_blend()
        >>> imgB = blend_images_multiply(img1, img2, alpha)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(imgB)
        >>> ut.show_if_requested()

    Ignore:
        >>> # GRIDSEARCH
        >>> from vtool_ibeis.blend import *  # NOQA
        >>> test_func = blend_images_multiply
        >>> args = testdata_blend(scale=128)
        >>> param_info = ut.ParamInfoList('blend_params', [
        ...    ut.ParamInfo('alpha', .8, 'alpha=',
        ...                 varyvals=np.linspace(0, 1.0, 9).tolist()),
        ... ])
        >>> gridsearch_image_function(param_info, test_func, args)
        >>> ut.show_if_requested()
    """
    import vtool_ibeis as vt
    # rectify type
    img1_ = vt.rectify_to_float01(img1)
    img2_ = vt.rectify_to_float01(img2)

    img1_, img2_ = vt.make_channels_comparable(img1_, img2_)

    # print(ub.repr2(ut.get_stats(img1_, axis=2)))
    # print(ub.repr2(ut.get_stats(img2_.ravel())))
    #assert img1_.min() >= 0 and img1_.max() <= 1
    #assert img2_.min() >= 0 and img2_.max() <= 1
    # apply transform
    #if False and alpha == .5:
    #imgB = img1_ * img2_
    #else:
    #data = [img1_, img2_]
    w1 = 1.0 - alpha + .5
    w2 = alpha + .5

    # w1 = alpha
    # w2 = (1 - alpha)

    #weights = [w1, w2]
    #imgB = vt.weighted_geometic_mean(data, weights)
    #imgB = ((img1_ ** w1) * (img2_ ** w2)) ** (1 / (w1 + w2))
    imgB = ((img1_ ** w1) * (img2_ ** w2))
    #imgB = vt.weighted_geometic_mean_unnormalized(data, weights)
    # unrectify
    #assert imgB.min() >= 0 and imgB.max() <= 1
    return imgB


def gridsearch_addWeighted():
    r"""
    CommandLine:
        xdoctest -m ~/code/vtool_ibeis/vtool_ibeis/blend.py gridsearch_addWeighted
    """
    import cv2
    import vtool_ibeis as vt
    def test_func(src1, src2, alpha=1.0, **kwargs):
        beta = 1.0 - alpha
        src1 = vt.rectify_to_float01(src1)
        src2 = vt.rectify_to_float01(src2)
        dst = np.empty(src1.shape, dtype=src1.dtype)
        cv2.addWeighted(src1=src1, src2=src2, dst=dst, alpha=alpha, beta=beta,
                        dtype=-1, **kwargs)
        return dst
    img1, img2 = testdata_blend()
    args = img1, img2 = vt.make_channels_comparable(img1, img2)
    param_info = ut.ParamInfoList('blend_params', [
        ut.ParamInfo('alpha', .8,
                     varyvals=np.linspace(0, 1.0, 5).tolist()),
        #ut.ParamInfo('beta', .8,
        ut.ParamInfo('gamma', .0,
                     varyvals=np.linspace(0, 1.0, 5).tolist()),
        #varyvals=[.0],))
        #ut.ParamInfo('gamma', .8, 'alpha=',
        #             varyvals=np.linspace(0, 1.0, 9).tolist()),
    ])
    gridsearch_image_function(param_info, test_func, args)


def gamma_adjust(img, gamma=1.0):
    """
    CommandLine:
        python -m vtool_ibeis.blend --test-gamma_adjust:0 --show

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.blend import *  # NOQA
        >>> import vtool_ibeis as vt
        >>> test_func = gamma_adjust
        >>> img_fpath = ut.grab_test_imgpath('carl.jpg')
        >>> img = vt.rectify_to_float01(vt.imread(img_fpath))
        >>> args = (img,)
        >>> param_info = ut.ParamInfoList('blend_params', [
        ...    ut.ParamInfo('gamma', .8, 'gamma=',
        ...                 varyvals=np.linspace(.1, 2.5, 25).tolist()),
        ... ])
        >>> gridsearch_image_function(param_info, test_func, args)
        >>> ut.show_if_requested()
    """
    assert img.max() <= 1.0
    assert img.min() >= 0.0
    return img ** gamma


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.blend
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



# Needs freetype-py>=1.0
# The MIT License (MIT)
#
# Copyright (c) 2013 Daniel Bader (http://dbader.org)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
References:
    https://dbader.org/blog/monochrome-font-rendering-with-freetype-and-python
    https://gist.github.com/dbader/5488053#file-fontdemo-py-L244
"""


class Bitmap(object):
    """
    A 2D bitmap image represented as a list of byte values. Each byte indicates the state
    of a single pixel in the bitmap. A value of 0 indicates that the pixel is `off`
    and any other value indicates that it is `on`.
    """
    def __init__(self, width, height, pixels=None):
        self.width = width
        self.height = height
        self.pixels = pixels or bytearray(width * height)

    def __repr__(self):
        """Return a string representation of the bitmap's pixels."""
        rows = ''
        for y in range(self.height):
            for x in range(self.width):
                rows += '#' if self.pixels[y * self.width + x] else '.'
            rows += '\n'
        return rows

    def bitblt(self, src, x, y):
        """Copy all pixels from `src` into this bitmap"""
        srcpixel = 0
        dstpixel = y * self.width + x
        row_offset = self.width - src.width
        for sy in range(src.height):
            for sx in range(src.width):
                self.pixels[dstpixel] = self.pixels[dstpixel] or src.pixels[srcpixel]
                srcpixel += 1
                dstpixel += 1
            dstpixel += row_offset


class Glyph(object):
    def __init__(self, pixels, width, height, top, advance_width):
        self.bitmap = Bitmap(width, height, pixels)
        self.top = top
        self.descent = max(0, self.height - self.top)
        self.ascent = max(0, max(self.top, self.height) - self.descent)
        self.advance_width = advance_width

    @property
    def width(self):
        return self.bitmap.width

    @property
    def height(self):
        return self.bitmap.height

    @staticmethod
    def from_glyphslot(slot):
        """Construct and return a Glyph object from a FreeType GlyphSlot."""
        pixels = Glyph.unpack_mono_bitmap(slot.bitmap)
        width, height = slot.bitmap.width, slot.bitmap.rows
        top = slot.bitmap_top
        advance_width = slot.advance.x // 64
        return Glyph(pixels, width, height, top, advance_width)

    @staticmethod
    def unpack_mono_bitmap(bitmap):
        """
        Unpack a freetype FT_LOAD_TARGET_MONO glyph bitmap into a bytearray where each
        pixel is represented by a single byte.
        """
        data = bytearray(bitmap.rows * bitmap.width)
        for y in range(bitmap.rows):
            for byte_index in range(bitmap.pitch):
                byte_value = bitmap.buffer[y * bitmap.pitch + byte_index]
                num_bits_done = byte_index * 8
                rowstart = y * bitmap.width + byte_index * 8
                for bit_index in range(min(8, bitmap.width - num_bits_done)):
                    bit = byte_value & (1 << (7 - bit_index))
                    data[rowstart + bit_index] = 1 if bit else 0
        return data


class Font(object):
    def __init__(self, filename, size):
        import freetype
        self.face = freetype.Face(filename)
        self.face.set_pixel_sizes(0, size)

    def glyph_for_character(self, char):
        import freetype
        self.face.load_char(char, freetype.FT_LOAD_RENDER | freetype.FT_LOAD_TARGET_MONO)
        return Glyph.from_glyphslot(self.face.glyph)

    def render_character(self, char):
        glyph = self.glyph_for_character(char)
        return glyph.bitmap

    def kerning_offset(self, previous_char, char):
        kerning = self.face.get_kerning(previous_char, char)
        return kerning.x // 64

    def text_dimensions(self, text):
        """
        Return (width, height, baseline) of `text` rendered in the current
        font. """
        width = 0
        max_ascent = 0
        max_descent = 0
        previous_char = None
        for char in text:
            glyph = self.glyph_for_character(char)
            max_ascent = max(max_ascent, glyph.ascent)
            max_descent = max(max_descent, glyph.descent)
            kerning_x = self.kerning_offset(previous_char, char)
            width += max(glyph.advance_width + kerning_x, glyph.width + kerning_x)
            previous_char = char
        height = max_ascent + max_descent
        return (width, height, max_descent)

    def render_text(self, text, width=None, height=None, baseline=None):
        """
        Render the given `text` into a Bitmap and return it.  If `width`,
        `height`, and `baseline` are not specified they are computed using the
        `text_dimensions' method.
        """
        if None in (width, height, baseline):
            width, height, baseline = self.text_dimensions(text)
        x = 0
        previous_char = None
        outbuffer = Bitmap(width, height)
        for char in text:
            glyph = self.glyph_for_character(char)
            x += self.kerning_offset(previous_char, char)
            y = height - glyph.ascent - baseline
            outbuffer.bitblt(glyph.bitmap, x, y)
            x += glyph.advance_width
            previous_char = char
        return outbuffer


def font_demo():
    r"""
    CommandLine:
        python -m vtool_ibeis.fontdemo font_demo --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.fontdemo import *  # NOQA
        >>> result = font_demo()
        >>> import utool as ut
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> ut.show_if_requested()
    """
    filename = '/usr/share/fonts/truetype/freefont/FreeMono.ttf'
    fnt = Font(filename, 12)
    # Single characters
    ch = fnt.render_character('e')
    print(repr(ch))
    # Multiple characters
    txt = fnt.render_text('hello')
    print(repr(txt))
    # Kerning
    print(repr(fnt.render_text('A1321')))
    # Choosing the baseline correctly
    print(repr(fnt.render_text('hello world')))


def get_text_test_img(text):
    r"""
    Args:
        text (str):

    CommandLine:
        python -m vtool_ibeis.fontdemo get_text_test_img --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from vtool_ibeis.fontdemo import *  # NOQA
        >>> import utool as ut
        >>> text = 'A012'
        >>> text_img = get_text_test_img(text)
        >>> result = ('text_img = %s' % (ub.repr2(text_img),))
        >>> print(result)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(text_img)
        >>> ut.show_if_requested()
    """
    filename = '/usr/share/fonts/truetype/freefont/FreeMono.ttf'
    fnt = Font(filename, 24)
    buf = fnt.render_text(text)
    import numpy as np
    img = np.array(buf.pixels)
    text_img = img.reshape(buf.height, buf.width) * 255
    return text_img

if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m vtool_ibeis.fontdemo
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
